from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Mapping, Sequence

    import numpy

    from .array_metadata import ArrayMetadataLike


class DeviceType(Enum):
    """An enum representing a device's type."""

    CPU = 1
    "CPU type"

    GPU = 2
    "GPU type"


class APIID:
    """An ID of an :py:class:`~grunnur.API` object."""

    shortcut: str
    """This API's shortcut."""

    def __init__(self, shortcut: str):
        self.shortcut = shortcut

    def __eq__(self, other: object) -> bool:
        return isinstance(other, APIID) and self.shortcut == other.shortcut

    def __hash__(self) -> int:
        return hash((type(self), self.shortcut))

    def __str__(self) -> str:
        return f"id({self.shortcut})"


class APIAdapterFactory(ABC):
    """
    A helper class that allows handling cases when an API's backend is unavailable
    or temporarily replaced by a mock object.
    """

    @property
    @abstractmethod
    def api_id(self) -> APIID:
        pass

    @property
    @abstractmethod
    def available(self) -> bool:
        pass

    @abstractmethod
    def make_api_adapter(self) -> APIAdapter:
        pass


class APIAdapter(ABC):
    @property
    @abstractmethod
    def id(self) -> APIID:
        pass

    @property
    @abstractmethod
    def platform_count(self) -> int:
        pass

    @abstractmethod
    def get_platform_adapters(self) -> tuple[PlatformAdapter, ...]:
        pass

    @abstractmethod
    def isa_backend_device(self, obj: Any) -> bool:
        pass

    @abstractmethod
    def isa_backend_platform(self, obj: Any) -> bool:
        pass

    @abstractmethod
    def isa_backend_context(self, obj: Any) -> bool:
        pass

    @abstractmethod
    def make_device_adapter(self, backend_device: Any) -> DeviceAdapter:
        pass

    @abstractmethod
    def make_platform_adapter(self, backend_platform: Any) -> PlatformAdapter:
        pass

    @abstractmethod
    def make_context_adapter_from_device_adapters(
        self, device_adapters: Sequence[DeviceAdapter]
    ) -> ContextAdapter:
        pass

    @abstractmethod
    def make_context_adapter_from_backend_contexts(
        self, backend_contexts: Sequence[Any], *, take_ownership: bool
    ) -> ContextAdapter:
        pass

    def __eq__(self, other: object) -> bool:
        return isinstance(other, APIAdapter) and self.id == other.id

    def __hash__(self) -> int:
        return hash((type(self), self.id))


class PlatformAdapter(ABC):
    @property
    @abstractmethod
    def api_adapter(self) -> APIAdapter:
        pass

    @property
    @abstractmethod
    def platform_idx(self) -> int:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def vendor(self) -> str:
        pass

    @property
    @abstractmethod
    def version(self) -> str:
        pass

    @property
    @abstractmethod
    def device_count(self) -> int:
        pass

    @abstractmethod
    def get_device_adapters(self) -> tuple[DeviceAdapter, ...]:
        pass


class DeviceAdapter(ABC):
    @property
    @abstractmethod
    def platform_adapter(self) -> PlatformAdapter:
        pass

    @property
    @abstractmethod
    def device_idx(self) -> int:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def params(self) -> DeviceParameters:
        pass


class DeviceParameters(ABC):
    """An object containing device's specifications."""

    @property
    @abstractmethod
    def type(self) -> DeviceType:
        """Device type."""

    @property
    @abstractmethod
    def max_total_local_size(self) -> int:
        """
        The maximum total number of threads in one block (CUDA),
        or work items in one work group (OpenCL).
        """

    @property
    @abstractmethod
    def max_local_sizes(self) -> tuple[int, ...]:
        """
        The maximum number of threads in one block (CUDA),
        or work items in one work group (OpenCL) for each of the available dimensions.
        """

    @property
    @abstractmethod
    def warp_size(self) -> int:
        """
        The number of threads (CUDA)/work items (OpenCL) that are executed synchronously
        (within one multiprocessor/compute unit).
        """

    @property
    @abstractmethod
    def max_num_groups(self) -> tuple[int, ...]:
        """
        The maximum number of blocks (CUDA)/work groups (OpenCL)
        for each of the available dimensions.
        """

    @property
    @abstractmethod
    def local_mem_size(self) -> int:
        """The size of shared (CUDA)/local (OpenCL) memory (in bytes)."""

    @property
    @abstractmethod
    def local_mem_banks(self) -> int:
        """
        The number of independent channels for shared (CUDA)/local (OpenCL) memory,
        which can be used from one warp without request serialization.
        """

    @property
    @abstractmethod
    def compute_units(self) -> int:
        """The number of multiprocessors (CUDA)/compute units (OpenCL) for the device."""

    @abstractmethod
    def align_words(self, word_size: int) -> int:
        """
        The distance (in ``word_size`` units) between global memory base addresses
        that allow accesses of word-size ``word_size`` bytes to get coalesced.
        """


class ContextAdapter(ABC):
    @property
    @abstractmethod
    def device_adapters(self) -> Mapping[int, DeviceAdapter]:
        pass

    @property
    @abstractmethod
    def device_order(self) -> list[int]:
        pass

    @abstractmethod
    def make_queue_adapter(self, device_adapter: DeviceAdapter) -> QueueAdapter:
        pass

    @abstractmethod
    def allocate(self, device_adapter: DeviceAdapter, size: int) -> BufferAdapter:
        pass

    @staticmethod
    @abstractmethod
    def render_prelude(*, fast_math: bool = False) -> str:
        """
        Renders the prelude allowing one to write kernels compiling
        both in CUDA and OpenCL.

        :param fast_math: whether the compilation with fast math is requested.
        """

    @abstractmethod
    def compile_single_device(
        self,
        device_adapter: DeviceAdapter,
        prelude: str,
        src: str,
        *,
        keep: bool = False,
        fast_math: bool = False,
        compiler_options: Sequence[str] | None = None,
        constant_arrays: Mapping[str, ArrayMetadataLike] | None = None,
    ) -> ProgramAdapter:
        """
        Compiles the given source with the given prelude on a single device.

        :param device_idx: the number of the device to compile on.
        :param prelude: the source of the prelude to prepend to the main source.
        :param src: the source of the kernels to be compiled.
        :param keep: see :py:meth:`compile`.
        :param fast_math: see :py:meth:`compile`.
        :param compiler_options: see :py:meth:`compile`.
        :param constant_arrays: (**CUDA only**) see :py:meth:`compile`.
        """

    @abstractmethod
    def deactivate(self) -> None:
        """
        For CUDA API: deactivates this context, popping all the CUDA context objects from the stack.
        Other APIs: no effect.
        """


class BufferAdapter(ABC):
    """A memory buffer on the device."""

    @property
    @abstractmethod
    def kernel_arg(self) -> Any:
        pass

    @property
    @abstractmethod
    def size(self) -> int:
        """The buffer's size (in bytes)."""

    @property
    @abstractmethod
    def offset(self) -> int:
        """
        The buffer's offset from the start of the physical memory allocation
        (will be non-zero for buffers created using :py:meth:`get_sub_region`).
        """

    @abstractmethod
    def get_sub_region(self, origin: int, size: int) -> BufferAdapter:
        """Returns a buffer sub-region starting at ``origin`` and of length ``size`` (in bytes)."""

    @abstractmethod
    def set(
        self,
        queue_adapter: QueueAdapter,
        source: numpy.ndarray[Any, numpy.dtype[Any]] | BufferAdapter,
        *,
        no_async: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def get(
        self,
        queue_adapter: QueueAdapter,
        host_array: numpy.ndarray[Any, numpy.dtype[Any]],
        *,
        async_: bool = False,
    ) -> None:
        pass


class QueueAdapter(ABC):
    @abstractmethod
    def synchronize(self) -> None:
        pass


class AdapterCompilationError(RuntimeError):
    def __init__(self, backend_exception: Exception, source: str):
        super().__init__(str(backend_exception))
        self.backend_exception = backend_exception
        self.source = source


class ProgramAdapter(ABC):
    @abstractmethod
    def __getattr__(self, kernel_name: str) -> KernelAdapter:
        pass

    @abstractmethod
    def set_constant_buffer(
        self,
        queue_adapter: QueueAdapter,
        name: str,
        arr: BufferAdapter | numpy.ndarray[Any, numpy.dtype[Any]],
    ) -> None:
        pass

    @property
    @abstractmethod
    def source(self) -> str:
        pass


class KernelAdapter(ABC):
    @property
    @abstractmethod
    def program_adapter(self) -> ProgramAdapter:
        pass

    @property
    @abstractmethod
    def max_total_local_size(self) -> int:
        pass

    @abstractmethod
    def prepare(
        self, global_size: Sequence[int], local_size: Sequence[int] | None = None
    ) -> PreparedKernelAdapter:
        pass


class PreparedKernelAdapter(ABC):
    @abstractmethod
    def __call__(
        self,
        queue_adapter: QueueAdapter,
        *args: BufferAdapter | numpy.generic,
        local_mem: int = 0,
    ) -> Any:
        pass



import pytest

from grunnur import Context, MultiQueue, Queue


def test_queue(mock_or_real_context):
    context, _mocked = mock_or_real_context
    queue = Queue(context.device)
    assert queue.device == context.devices[0]

    queue.synchronize()


def test_queue_on_multi_device_context(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    queue = Queue(context.devices[1])
    assert queue.device == context.devices[1]


def test_multi_queue(mock_or_real_multi_device_context):
    context, _mocked = mock_or_real_multi_device_context
    mqueue = MultiQueue.on_devices(context.devices)
    assert set(mqueue.queues.keys()) == {context.devices[0], context.devices[1]}
    assert mqueue.devices == context.devices[[0, 1]]

    mqueue.synchronize()

    mqueue = MultiQueue.on_devices(context.devices[[1]])
    assert set(mqueue.queues.keys()) == {context.devices[1]}
    assert mqueue.devices == context.devices[[1]]

    mqueue.synchronize()


def test_multi_queue_out_of_queues(mock_4_device_context):
    context = mock_4_device_context
    queue0 = Queue(context.devices[0])
    queue1 = Queue(context.devices[1])

    mqueue = MultiQueue([queue0, queue1])
    assert set(mqueue.queues.keys()) == {context.devices[0], context.devices[1]}
    assert mqueue.devices == context.devices[[0, 1]]



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, NamedTuple

from .api import API
from .platform import Platform
from .utils import string_matches_masks

if TYPE_CHECKING:  # pragma: no cover
    from .adapter_base import DeviceAdapter, DeviceParameters


class DeviceFilter(NamedTuple):
    """A set of filters for device discovery."""

    include_masks: list[str] | None = None
    """A list of strings (treated as regexes), one of which must match the device name."""

    exclude_masks: list[str] | None = None
    """A list of strings (treated as regexes), neither of which must match the device name."""

    unique_only: bool = False
    """If ``True``, only return devices with unique names."""

    exclude_pure_parallel: bool = False
    """
    If ``True``, exclude devices with ``max_total_local_size`` equal to 1
    (that is the ones that do not support workgroup synchronization).
    """


class Device:
    """A generalized GPGPU device."""

    platform: Platform
    """The :py:class:`~grunnur.Platform` object this device belongs to."""

    name: str
    """This device's name."""

    @classmethod
    def all(cls, platform: Platform) -> list[Device]:
        """
        Returns a list of devices available for the given platform.

        :param platform: the platform to search in.
        """
        return [
            Device.from_index(platform, device_idx)
            for device_idx in range(platform._platform_adapter.device_count)  # noqa: SLF001
        ]

    @classmethod
    def all_filtered(
        cls,
        platform: Platform,
        filter_: DeviceFilter | None = None,
    ) -> list[Device]:
        """
        Returns a list of all devices satisfying the given criteria in the given platform.
        If ``filter`` is not provided, returns all the devices.
        """
        if filter_ is None:
            return cls.all(platform)

        seen_devices = set()
        devices = []

        for device in cls.all(platform):
            if not string_matches_masks(
                device.name,
                include_masks=filter_.include_masks,
                exclude_masks=filter_.exclude_masks,
            ):
                continue

            if filter_.unique_only and device.name in seen_devices:
                continue

            if filter_.exclude_pure_parallel and device.params.max_total_local_size == 1:
                continue

            seen_devices.add(device.name)
            devices.append(device)

        return devices

    @classmethod
    def from_backend_device(cls, obj: Any) -> Device:
        """Wraps a backend device object into a Grunnur device object."""
        for api in API.all_available():
            if api._api_adapter.isa_backend_device(obj):  # noqa: SLF001
                device_adapter = api._api_adapter.make_device_adapter(obj)  # noqa: SLF001
                return cls(device_adapter)

        raise TypeError(f"{obj} was not recognized as a device object by any available API")

    @classmethod
    def from_index(cls, platform: Platform, device_idx: int) -> Device:
        """
        Creates a device based on its index in the list returned by the API.

        :param platform: the API to search in.
        :param device_idx: the target device's index.
        """
        device_adapter = platform._platform_adapter.get_device_adapters()[device_idx]  # noqa: SLF001
        return cls(device_adapter)

    def __init__(self, device_adapter: DeviceAdapter):
        self.platform = Platform(device_adapter.platform_adapter)
        self._device_adapter = device_adapter
        self.name = self._device_adapter.name

        self.shortcut = f"{self.platform.shortcut},{device_adapter.device_idx}"

        self._params = None

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Device) and self._device_adapter == other._device_adapter

    def __hash__(self) -> int:
        return hash((type(self), self._device_adapter))

    @property
    def params(self) -> DeviceParameters:
        """
        Returns a :py:class:`~grunnur.adapter_base.DeviceParameters` object
        associated with this device.
        """
        # Already cached in the adapters
        return self._device_adapter.params

    def __str__(self) -> str:
        return f"device({self.shortcut})"



import numpy
import pytest

from grunnur import (
    API,
    Array,
    Context,
    MultiArray,
    MultiQueue,
    Queue,
    StaticKernel,
    VirtualSizeError,
    cuda_api_id,
    opencl_api_id,
)
from grunnur.template import DefTemplate
from grunnur.testing import MockDefTemplate, MockKernel, PyCUDADeviceInfo, PyOpenCLDeviceInfo
from test_program import _test_constant_memory

SRC = """
KERNEL void multiply(GLOBAL_MEM int *dest, GLOBAL_MEM int *a, GLOBAL_MEM int *b)
{
    if (${static.skip}()) return;
    const int i = ${static.global_id}(0);
    const int j = ${static.global_id}(1);
    const int idx = ${static.global_flat_id}();
    dest[idx] = a[i] * b[j];
}
"""


def test_compile_static(mock_or_real_context):
    context, mocked = mock_or_real_context

    if mocked:
        kernel = MockKernel("multiply", [None, None, None], max_total_local_sizes={0: 1024})
        src = MockDefTemplate(kernels=[kernel])
    else:
        src = SRC

    a = numpy.arange(11).astype(numpy.int32)
    b = numpy.arange(15).astype(numpy.int32)
    ref = numpy.outer(a, b)

    queue = Queue(context.device)

    a_dev = Array.from_host(queue, a)
    b_dev = Array.from_host(queue, b)

    res_dev = Array.empty(context.device, (11, 15), numpy.int32)

    multiply = StaticKernel([context.device], src, "multiply", (11, 15))
    multiply(queue, res_dev, a_dev, b_dev)

    res = res_dev.get(queue)

    if not mocked:
        assert (res == ref).all()


def test_compile_static_multi_device(mock_or_real_multi_device_context):
    context, mocked = mock_or_real_multi_device_context

    if mocked:
        kernel = MockKernel("multiply", [None, None, None], max_total_local_sizes={0: 1024, 1: 512})
        src = MockDefTemplate(kernels=[kernel])
    else:
        src = SRC

    a = numpy.arange(22).astype(numpy.int32)
    b = numpy.arange(15).astype(numpy.int32)
    ref = numpy.outer(a, b)

    mqueue = MultiQueue.on_devices(context.devices[[0, 1]])

    a_dev = MultiArray.from_host(mqueue, a)
    b_dev = MultiArray.from_host(mqueue, b, splay=MultiArray.CloneSplay())
    res_dev = MultiArray.empty(mqueue.devices, (22, 15), ref.dtype)

    multiply = StaticKernel(mqueue.devices, src, "multiply", res_dev.shapes)
    multiply(mqueue, res_dev, a_dev, b_dev)

    res = res_dev.get(mqueue)

    if not mocked:
        assert (res == ref).all()


def test_constant_memory(mock_or_real_context):
    context, mocked = mock_or_real_context
    _test_constant_memory(context=context, mocked=mocked, is_static=True)


def test_find_local_size(mock_context):
    kernel = MockKernel("multiply", [None], max_total_local_sizes={0: 64})
    src = MockDefTemplate(kernels=[kernel])
    multiply = StaticKernel([mock_context.device], src, "multiply", (11, 15))
    assert multiply._vs_metadata[mock_context.devices[0]].real_global_size == (16, 12)
    assert multiply._vs_metadata[mock_context.devices[0]].real_local_size == (16, 4)


def test_reserved_names(mock_context):
    kernel = MockKernel("test", [None])
    src = MockDefTemplate(kernels=[kernel])
    with pytest.raises(ValueError, match="The global name 'static' is reserved in static kernels"):
        StaticKernel([mock_context.device], src, "test", (1024,), render_globals=dict(static=1))


def test_zero_max_total_local_size(mock_context):
    kernel = MockKernel("test", [None], max_total_local_sizes={0: 0})
    src = MockDefTemplate(kernels=[kernel])
    with pytest.raises(
        VirtualSizeError,
        match="The kernel requires too much resourses to be executed with any local size",
    ):
        StaticKernel([mock_context.device], src, "test", (1024,))


def test_virtual_sizes_error_propagated(mock_backend_pycuda):
    # Testing for PyCUDA backend only since mocked PyOpenCL backend does not have a way
    # to set maximum global sizes (PyOpenCL devices don't have a corresponding parameter),
    # and PyCUDA is enough to test the required code path.

    device_info = PyCUDADeviceInfo(
        max_threads_per_block=2**4,
        max_block_dim_x=2**4,
        max_block_dim_y=2**4,
        max_block_dim_z=2**4,
        max_grid_dim_x=2**10,
        max_grid_dim_y=2**10,
        max_grid_dim_z=2**8,
    )

    mock_backend_pycuda.add_devices([device_info])
    api = API.from_api_id(mock_backend_pycuda.api_id)
    device = api.platforms[0].devices[0]
    context = Context.from_devices([device])
    kernel = MockKernel("test", [None], max_total_local_sizes={0: 16})
    src = MockDefTemplate(kernels=[kernel])

    # Just enough to fit in the grid limits
    StaticKernel([context.device], src, "test", (2**14, 2**10, 2**8), local_size=(2**4, 1, 1))

    # Global size is too large to fit on the device,
    # so virtual size finding fails and the error is propagated to the user.
    with pytest.raises(
        VirtualSizeError, match="Bounding global size \\[16384, 2048, 256\\] is too large"
    ):
        StaticKernel([context.device], src, "test", (2**14, 2**11, 2**8), local_size=(2**4, 1, 1))


def test_builtin_globals(mock_backend_pycuda):
    mock_backend_pycuda.add_devices(
        [PyCUDADeviceInfo(max_threads_per_block=1024), PyCUDADeviceInfo(max_threads_per_block=512)]
    )

    source_template = DefTemplate.from_string(
        "mock_source",
        [],
        """
        KERNEL void test()
        {
            int max_total_local_size = ${device_params.max_total_local_size};
        }
        """,
    )

    api = API.from_api_id(mock_backend_pycuda.api_id)
    context = Context.from_devices([api.platforms[0].devices[0], api.platforms[0].devices[1]])

    src = MockDefTemplate(
        kernels=[MockKernel("test", [None], max_total_local_sizes={0: 1024, 1: 512})],
        source_template=source_template,
    )

    kernel = StaticKernel(context.devices, src, "test", (1024,))

    assert "max_total_local_size = 1024" in kernel.sources[context.devices[0]].source
    assert "max_total_local_size = 512" in kernel.sources[context.devices[1]].source



from typing import Any, Protocol

from .. import all_api_ids, cuda_api_id, opencl_api_id
from ..adapter_base import APIID
from .mock_pycuda import MockPyCUDA
from .mock_pyopencl import MockPyOpenCL


# Have to mock it like this since we don't want to make `pytest` a dependency for the main library.
class MonkeyPatch(Protocol):
    def setattr(self, attr: str, obj: Any) -> None:  # pragma: no cover
        ...


class MockBackendFactory:
    def __init__(self, monkeypatch: MonkeyPatch):
        self.monkeypatch = monkeypatch

        # Clear out all existing backends
        for api_id in all_api_ids():
            self.disable(api_id)

    def _set_backend_cuda(self, backend: MockPyCUDA | None = None) -> None:
        pycuda_driver = backend.pycuda_driver if backend else None
        pycuda_compiler = backend.pycuda_compiler if backend else None
        pycuda_tools = backend.pycuda_tools if backend else None
        self.monkeypatch.setattr("grunnur.adapter_cuda.pycuda_driver", pycuda_driver)
        self.monkeypatch.setattr("grunnur.adapter_cuda.pycuda_compiler", pycuda_compiler)
        self.monkeypatch.setattr("grunnur.adapter_cuda.pycuda_tools", pycuda_tools)

    def mock_pycuda(self) -> MockPyCUDA:
        backend = MockPyCUDA()
        self._set_backend_cuda(backend)
        return backend

    def disable_pycuda(self) -> None:
        self._set_backend_cuda(None)

    def _set_backend_opencl(self, backend: MockPyOpenCL | None = None) -> None:
        pyopencl = backend.pyopencl if backend else None
        self.monkeypatch.setattr("grunnur.adapter_opencl.pyopencl", pyopencl)

    def mock_pyopencl(self) -> MockPyOpenCL:
        backend = MockPyOpenCL()
        self._set_backend_opencl(backend)
        return backend

    def disable_pyopencl(self) -> None:
        self._set_backend_opencl(None)

    def disable(self, api_id: APIID) -> None:
        if api_id == cuda_api_id():
            return self.disable_pycuda()
        if api_id == opencl_api_id():
            return self.disable_pyopencl()
        raise ValueError(f"Unknown API ID: {api_id}")

    def mock(self, api_id: APIID) -> MockPyOpenCL | MockPyCUDA:
        if api_id == cuda_api_id():
            return self.mock_pycuda()
        if api_id == opencl_api_id():
            return self.mock_pyopencl()
        raise ValueError(f"Unknown API ID: {api_id}")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, cast, runtime_checkable

import numpy

from .array_metadata import ArrayMetadata, ArrayMetadataLike
from .buffer import Buffer
from .context import BoundDevice, BoundMultiDevice, Context
from .device import Device
from .queue import MultiQueue, Queue
from .utils import min_blocks

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable, Mapping, Sequence

    from numpy.typing import DTypeLike


class Array:
    """Array on a single device."""

    device: BoundDevice
    """Device this array is allocated on."""

    shape: tuple[int, ...]
    """Array shape."""

    dtype: numpy.dtype[Any]
    """Array item data type."""

    strides: tuple[int, ...]
    """Array strides."""

    @classmethod
    def from_host(
        cls,
        queue_or_device: Queue | BoundDevice,
        host_arr: numpy.ndarray[Any, numpy.dtype[Any]],
    ) -> Array:
        """
        Creates an array object from a host array.

        :param queue_or_device: the queue to use for the transfer, or the target device.
            In the latter case an operation will be performed synchronously.
        :param host_arr: the source array.
        """
        if isinstance(queue_or_device, BoundDevice):
            queue = Queue(queue_or_device)
        else:
            queue = queue_or_device
        array = cls.empty(queue.device, host_arr.shape, host_arr.dtype)
        array.set(queue, host_arr)
        if isinstance(queue_or_device, BoundDevice):
            queue.synchronize()
        return array

    @classmethod
    def empty(
        cls,
        device: BoundDevice,
        shape: Sequence[int],
        dtype: DTypeLike,
        strides: Sequence[int] | None = None,
        first_element_offset: int = 0,
        allocator: Callable[[BoundDevice, int], Buffer] | None = None,
    ) -> Array:
        """
        Creates an empty array.

        :param device: device on which this array will be allocated.
        :param shape: array shape.
        :param dtype: array data type.
        :param allocator: an optional callable taking two arguments
            (the bound device, and the buffer size in bytes)
            and returning a :py:class:`Buffer` object.
            If ``None``, will use :py:meth:`Buffer.allocate`.
        """
        metadata = ArrayMetadata(
            shape, dtype, strides=strides, first_element_offset=first_element_offset
        )
        size = metadata.buffer_size
        if allocator is None:
            allocator = Buffer.allocate
        data = allocator(device, size)

        return cls(metadata, data)

    @classmethod
    def empty_like(cls, device: BoundDevice, array_like: ArrayMetadataLike) -> Array:
        """Creates an empty array with the same shape and dtype as ``array_like``."""
        return cls.empty(device, array_like.shape, array_like.dtype)

    def __init__(self, array_metadata: ArrayMetadata, data: Buffer):
        self._metadata = array_metadata

        self.device = data.device
        self.shape = self._metadata.shape
        self.dtype = self._metadata.dtype
        self.strides = self._metadata.strides
        self.first_element_offset = self._metadata.first_element_offset
        self.buffer_size = self._metadata.buffer_size

        if data.size < self.buffer_size:
            raise ValueError(
                "Provided data buffer is not big enough to hold the array "
                "(minimum required {self.buffer_size})"
            )

        self.data = data

    def _view(self, slices: slice | tuple[slice, ...]) -> Array:
        new_metadata = self._metadata[slices]

        origin, size, new_metadata = new_metadata.minimal_subregion()
        data = self.data.get_sub_region(origin, size)
        return Array(new_metadata, data)

    def set(
        self,
        queue: Queue,
        array: numpy.ndarray[Any, numpy.dtype[Any]] | Array,
        *,
        no_async: bool = False,
    ) -> None:
        """
        Copies the contents of the host array to the array.

        :param queue: the queue to use for the transfer.
        :param array: the source array.
        :param no_async: if `True`, the transfer blocks until completion.
        """
        array_data: numpy.ndarray[Any, numpy.dtype[Any]] | Buffer
        if isinstance(array, numpy.ndarray):
            array_data = array
        elif isinstance(array, Array):
            if not array._metadata.contiguous:  # noqa: SLF001
                raise ValueError("Setting from a non-contiguous device array is not supported")
            array_data = array.data
        else:
            raise TypeError(f"Cannot set from an object of type {type(array)}")

        if self.shape != array.shape:
            raise ValueError(f"Shape mismatch: expected {self.shape}, got {array.shape}")
        if self.dtype != array.dtype:
            raise ValueError(f"Dtype mismatch: expected {self.dtype}, got {array.dtype}")

        self.data.set(queue, array_data, no_async=no_async)

    def get(
        self,
        queue: Queue,
        dest: numpy.ndarray[Any, numpy.dtype[Any]] | None = None,
        *,
        async_: bool = False,
    ) -> numpy.ndarray[Any, numpy.dtype[Any]]:
        """
        Copies the contents of the array to the host array and returns it.

        :param queue: the queue to use for the transfer.
        :param dest: the destination array. If ``None``, the target array is created.
        :param async_: if `True`, the transfer is performed asynchronously.
        """
        if dest is None:
            dest = numpy.empty(self.shape, self.dtype)
        self.data.get(queue, dest, async_=async_)
        return dest

    def __getitem__(self, slices: slice | tuple[slice, ...]) -> Array:
        """Returns a view of this array."""
        return self._view(slices)


@runtime_checkable
class ArrayLike(ArrayMetadataLike, Protocol):
    """
    A protocol for an array-like object supporting views via ``__getitem__()``.
    :py:class:`numpy.ndarray` or :py:class:`Array` follow this protocol.
    """

    def __getitem__(
        self: _ArrayLike, slices: slice | tuple[slice, ...]
    ) -> _ArrayLike:  # pragma: no cover
        """Returns a view of this array."""
        ...


_ArrayLike = TypeVar("_ArrayLike", bound=ArrayLike)
"""Any type that follows the :py:class:`ArrayLike` protocol."""


class BaseSplay(ABC):
    """Base class for splay strategies for :py:class:`~grunnur.MultiArray`."""

    @abstractmethod
    def __call__(
        self, arr: _ArrayLike, devices: Sequence[BoundDevice]
    ) -> dict[BoundDevice, _ArrayLike]:
        """
        Creates a dictionary of views of an array-like object for each of the given devices.

        :param arr: an array-like object.
        :param devices: a multi-device object.
        """


class EqualSplay(BaseSplay):
    """
    Splays the given array equally between the devices using the outermost dimension.
    The outermost dimension should be larger or equal to the number of devices.
    """

    def __call__(
        self, arr: _ArrayLike, devices: Sequence[BoundDevice]
    ) -> dict[BoundDevice, _ArrayLike]:
        parts = len(devices)
        outer_dim = arr.shape[0]
        if parts > outer_dim:
            raise ValueError(
                "The number of devices to splay to "
                "cannot be greater than the outer array dimension"
            )
        chunk_size = min_blocks(outer_dim, parts)
        rem = outer_dim % parts

        subarrays = {}
        ptr = 0
        for part, device in enumerate(devices):
            length = chunk_size if rem == 0 or part < rem else chunk_size - 1
            subarrays[device] = arr[ptr : ptr + length]
            ptr += length

        return subarrays


class CloneSplay(BaseSplay):
    """Copies the given array to each device."""

    def __call__(
        self, arr: _ArrayLike, devices: Sequence[BoundDevice]
    ) -> dict[BoundDevice, _ArrayLike]:
        return {device: arr for device in devices}


class MultiArray:
    """An array on multiple devices."""

    devices: BoundMultiDevice
    """Devices on which the sub-arrays are allocated"""

    shape: tuple[int, ...]
    """Array shape."""

    dtype: numpy.dtype[Any]
    """Array item data type."""

    shapes: dict[BoundDevice, tuple[int, ...]]
    """Sub-array shapes matched to device indices."""

    subarrays: dict[BoundDevice, Array]

    EqualSplay = EqualSplay
    CloneSplay = CloneSplay

    @classmethod
    def from_host(
        cls,
        mqueue: MultiQueue,
        host_arr: numpy.ndarray[Any, numpy.dtype[Any]],
        splay: BaseSplay | None = None,
    ) -> MultiArray:
        """
        Creates an array object from a host array.

        :param mqueue: the queue to use for the transfer.
        :param host_arr: the source array.
        :param splay: the splay strategy (if ``None``, an :py:class:`EqualSplay` object is used).
        """
        if splay is None:
            splay = EqualSplay()

        # This is to satisfy Mypy. Numpy arrays satisfy this protocol by default.
        assert isinstance(host_arr, ArrayMetadataLike)  # noqa: S101
        host_subarrays = splay(host_arr, list(mqueue.queues))

        subarrays = {
            device: Array.from_host(mqueue.queues[device], host_subarrays[device])
            for device in mqueue.devices
        }

        return cls(mqueue.devices, ArrayMetadata(host_arr.shape, host_arr.dtype), subarrays, splay)

    @classmethod
    def empty(
        cls,
        devices: BoundMultiDevice,
        shape: Sequence[int],
        dtype: DTypeLike,
        allocator: Callable[[BoundDevice, int], Buffer] | None = None,
        splay: BaseSplay | None = None,
    ) -> MultiArray:
        """
        Creates an empty array.

        :param devices: devices on which the sub-arrays will be allocated.
        :param shape: array shape.
        :param dtype: array data type.
        :param allocator: an optional callable taking two integer arguments
            (the device to allocate it on and the buffer size in bytes)
            and returning a :py:class:`Buffer` object.
            If ``None``, will use :py:meth:`Buffer.allocate`.
        :param splay: the splay strategy (if ``None``, an :py:class:`EqualSplay` object is used).
        """
        if splay is None:
            splay = EqualSplay()

        metadata = ArrayMetadata(shape, dtype)
        submetadatas = splay(metadata, devices)

        subarrays = {
            device: Array.empty(device, submetadata.shape, submetadata.dtype, allocator=allocator)
            for device, submetadata in submetadatas.items()
        }

        return cls(devices, metadata, subarrays, splay)

    def __init__(
        self,
        devices: BoundMultiDevice,
        metadata: ArrayMetadata,
        subarrays: Mapping[BoundDevice, Array],
        splay: BaseSplay,
    ):
        self.devices = devices
        self.shape = metadata.shape
        self.dtype = metadata.dtype
        self.subarrays = dict(subarrays)
        self.shapes = {device: subarray.shape for device, subarray in self.subarrays.items()}

        self._splay = splay

    def get(
        self,
        mqueue: MultiQueue,
        dest: numpy.ndarray[Any, numpy.dtype[Any]] | None = None,
        *,
        async_: bool = False,
    ) -> numpy.ndarray[Any, numpy.dtype[Any]]:
        """
        Copies the contents of the array to the host array and returns it.

        :param mqueue: the queue to use for the transfer.
        :param dest: the destination array. If ``None``, the target array is created.
        :param async_: if `True`, the transfer is performed asynchronously.
        """
        if dest is None:
            dest = numpy.empty(self.shape, self.dtype)

        # This is to satisfy Mypy. Numpy arrays satisfy this protocol by default.
        assert isinstance(dest, ArrayLike)  # noqa: S101
        dest_subarrays = self._splay(dest, list(self.subarrays))

        for device, subarray in self.subarrays.items():
            subarray.get(mqueue.queues[device], dest_subarrays[device], async_=async_)

        return dest

    def set(
        self,
        mqueue: MultiQueue,
        array: numpy.ndarray[Any, numpy.dtype[Any]] | MultiArray,
        *,
        no_async: bool = False,
    ) -> None:
        """
        Copies the contents of the host array to the array.

        :param mqueue: the queue to use for the transfer.
        :param array: the source array.
        :param no_async: if `True`, the transfer blocks until completion.
        """
        subarrays: Mapping[BoundDevice, Array | numpy.ndarray[Any, numpy.dtype[Any]]]
        if isinstance(array, numpy.ndarray):
            # This is to satisfy Mypy. Numpy arrays satisfy this protocol by default.
            assert isinstance(array, ArrayLike)  # noqa: S101
            subarrays = self._splay(array, self.devices)
        else:
            subarrays = array.subarrays

        if self.subarrays.keys() != subarrays.keys():
            raise ValueError("Mismatched device sets in the source and the destination")

        for device in self.subarrays:
            self.subarrays[device].set(mqueue.queues[device], subarrays[device], no_async=no_async)



"""
An example comparing performance of executing a kernel on one device
versus multiple devices simultaneously.
"""

# ruff: noqa: T201, S101

import time
from collections.abc import Sequence
from typing import Any

import numpy

from grunnur import Array, Context, MultiArray, MultiQueue, Program, Queue
from grunnur import opencl_api as api

src = """
KERNEL void sum(GLOBAL_MEM unsigned long *a, int pwr)
{
    unsigned long m = 4294967231;
    int gid = get_global_id(0);

    int idx = gid;
    unsigned long x = a[idx];
    unsigned long res = x;
    for (int i = 1; i < pwr; i++)
    {
        res *= x;
        res %= m;
    }
    a[idx] = res;
}
"""


def calc_ref(
    x: "numpy.ndarray[Any, numpy.dtype[Any]]", pwr: int
) -> "numpy.ndarray[Any, numpy.dtype[Any]]":
    """Reference function for the kernel."""
    m = numpy.uint64(2**32 - 65)
    res = x.copy()
    for _ in range(1, pwr):
        res *= x
        res %= m
    return res


def test_single_device(device_idx: int, full_len: int, *, benchmark: bool = False) -> None:
    """Runs the kernel on a single device."""
    pwr = 50

    a = numpy.arange(full_len).astype(numpy.uint64)

    context = Context.from_devices([api.platforms[0].devices[device_idx]])
    queue = Queue(context.device)

    program = Program([context.device], src)
    a_dev = Array.from_host(queue, a)

    queue.synchronize()
    t1 = time.time()
    program.kernel.sum(queue, [full_len], None, a_dev, numpy.int32(pwr))
    queue.synchronize()
    t2 = time.time()
    print(f"Single device time (device {device_idx}):", t2 - t1)

    a_res = a_dev.get(queue)

    if not benchmark:
        a_ref = calc_ref(a, pwr)
        assert (a_ref == a_res).all()


def test_multi_device(
    device_idxs: Sequence[int], full_len: int, *, benchmark: bool = False
) -> None:
    """Runs the kernel on a multiple devices."""
    pwr = 50

    a = numpy.arange(full_len).astype(numpy.uint64)

    context = Context.from_devices(
        [api.platforms[0].devices[device_idx] for device_idx in device_idxs]
    )
    mqueue = MultiQueue.on_devices(context.devices)

    program = Program(context.devices, src)
    a_dev = MultiArray.from_host(mqueue, a)

    mqueue.synchronize()
    t1 = time.time()
    program.kernel.sum(mqueue, a_dev.shapes, None, a_dev, numpy.int32(pwr))
    mqueue.synchronize()
    t2 = time.time()
    print(f"Multidevice time (devices {device_idxs}):", t2 - t1)

    a_res = a_dev.get(mqueue)

    if not benchmark:
        a_ref = calc_ref(a, pwr)
        assert (a_ref == a_res).all()


test_single_device(1, 2**20)
test_single_device(2, 2**20)
test_multi_device([1, 2], 2**20)

test_single_device(1, 2**24, benchmark=True)
test_single_device(2, 2**24, benchmark=True)
test_multi_device([1, 2], 2**24, benchmark=True)



import collections
import re
from collections.abc import Iterable, Mapping, Sequence
from functools import reduce
from typing import Any, TypeVar, overload

_T = TypeVar("_T")


def all_same(seq: Iterable[Any]) -> bool:
    seq = list(seq)
    return all(elem == seq[0] for elem in seq[1:])


def all_different(seq: Iterable[Any]) -> bool:
    seq = list(seq)
    return len(seq) == len(set(seq))


def min_blocks(length: int, block: int) -> int:
    """
    Returns minimum number of blocks with length ``block``
    necessary to cover the array with non-zero length ``length``.
    """
    if length <= 0:
        raise ValueError("The length must be positive")
    return (length - 1) // block + 1


def log2(num: int) -> int:
    """
    Integer-valued logarigthm with base 2.
    If ``n`` is not a power of 2, the result is rounded to the smallest number.
    """
    return num.bit_length() - 1


def bounding_power_of_2(num: int) -> int:
    """
    Returns the minimal number of the form ``2**m``
    such that it is greater or equal to ``n``.
    """
    if num == 1:
        return 1
    return 1 << (log2(num - 1) + 1)


def prod(seq: Iterable[int]) -> int:
    # Integer product. `numpy.prod` returns a float when given an empty sequence.
    return reduce(lambda x, y: x * y, seq, 1)


def string_matches_masks(
    s: str,
    include_masks: Sequence[str] | None = None,
    exclude_masks: Sequence[str] | None = None,
) -> bool:
    """
    Returns ``True`` if:
    1) ``s`` matches with at least one of regexps from ``include_masks`` (if there are any), and
    2) ``s`` doesn't match any regexp from ``exclude_masks``.
    """
    if include_masks is None:
        include_masks = []
    if exclude_masks is None:
        exclude_masks = []

    if len(include_masks) > 0:
        for include_mask in include_masks:
            if re.search(include_mask, s):
                break
        else:
            return False

    if len(exclude_masks) > 0:
        for exclude_mask in exclude_masks:
            if re.search(exclude_mask, s):
                return False

    return True


def normalize_object_sequence(objs: Sequence[Any], expected_cls: type[_T]) -> tuple[_T, ...]:
    """
    For a sequence of objects, or a single object, checks that:
    1) the sequence is non-empty;
    2) all objects are different; and
    3) every object is an instance of ``expected_cls``,
    raising a ``ValueError`` otherwise.
    Returns a tuple of objects (1-tuple, if there was a single object).
    """
    if len(objs) == 0:
        raise ValueError("The iterable of base objects for the context cannot be empty")

    if not all_different(objs):
        raise ValueError("All base objects must be different")

    types = [type(obj) for obj in objs]
    if not all(issubclass(tp, expected_cls) for tp in types):
        raise TypeError(f"The iterable must contain only subclasses of {expected_cls}, got {types}")

    return tuple(objs)


def max_factor(x: int, y: int) -> int:
    """Find the maximum `d` such that `x % d == 0` and `d <= y`."""
    if x <= y:
        return x

    result = 1
    for d in range(2, min(int(x**0.5), y) + 1):
        inv_d = x // d
        if inv_d * d == x:
            if inv_d <= y:
                return inv_d
            result = d

    return result


def find_local_size(
    global_size: Sequence[int], max_local_sizes: Sequence[int], max_total_local_size: int
) -> tuple[int, ...]:
    """
    Mimics the OpenCL local size finding algorithm.
    Returns the tuple of the same length as ``global_size``, with every element
    being a factor of the corresponding element of ``global_size``.
    Neither of the elements of ``local_size`` are greater then the corresponding element
    of ``max_local_sizes``, and their product is not greater than ``max_total_local_size``.
    """
    local_size = []
    for gs, mls in zip(global_size, max_local_sizes, strict=False):
        d = max_factor(gs, min(mls, max_total_local_size))
        max_total_local_size //= d
        local_size.append(d)

    return tuple(local_size)


def get_launch_size(
    max_local_sizes: tuple[int, ...],
    max_total_local_size: int,
    global_size: Sequence[int],
    local_size: Sequence[int] | None = None,
) -> tuple[tuple[int, ...], tuple[int, ...]]:
    """
    Constructs the grid and block tuples to launch a CUDA kernel
    based on the provided global and local sizes.
    """
    if len(global_size) > len(max_local_sizes):
        raise ValueError("Global size has too many dimensions")

    if local_size is not None:
        if len(local_size) != len(global_size):
            raise ValueError("Global/local work sizes have differing dimensions")
        for gs, ls in zip(global_size, local_size, strict=True):
            if gs % ls != 0:
                raise ValueError("Global sizes must be multiples of corresponding local sizes")
        local_size_tuple = tuple(local_size)
    else:
        local_size_tuple = find_local_size(global_size, max_local_sizes, max_total_local_size)

    grid_size = tuple(gs // ls for gs, ls in zip(global_size, local_size_tuple, strict=True))
    return grid_size, local_size_tuple


_UPDATE_ERROR_TEMPLATE = "Cannot add an item '{name}' - it already exists in the old dictionary"


def update_dict(
    d: Mapping[str, _T], new_d: Mapping[str, _T], error_msg: str = _UPDATE_ERROR_TEMPLATE
) -> dict[str, _T]:
    res = dict(d)
    for name, value in new_d.items():
        if name in d:
            raise ValueError(error_msg.format(name=name))
        res[name] = value
    return res



from __future__ import annotations

import functools
import itertools
import platform
from collections.abc import Callable, Iterable, Mapping, Sequence
from math import gcd
from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar, cast, overload

import numpy

from .modules import Module
from .utils import bounding_power_of_2, log2, min_blocks, prod

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import DTypeLike


_DTYPE_TO_BUILTIN_CTYPE = {
    numpy.dtype("bool"): "bool",
    numpy.dtype("int8"): "char",
    numpy.dtype("uint8"): "unsigned char",
    numpy.dtype("int16"): "short",
    numpy.dtype("uint16"): "unsigned short",
    numpy.dtype("int32"): "int",
    numpy.dtype("uint32"): "unsigned int",
    numpy.dtype("float32"): "float",
    numpy.dtype("float64"): "double",
    numpy.dtype("complex64"): "float2",
    numpy.dtype("complex128"): "double2",
    numpy.dtype("int64"): "long long" if platform.system() == "Windows" else "long",
    numpy.dtype("uint64"): "unsigned "
    + ("long long" if platform.system() == "Windows" else "long"),
}


def _ctype_builtin(dtype: numpy.dtype[Any]) -> str:
    if dtype in _DTYPE_TO_BUILTIN_CTYPE:
        return _DTYPE_TO_BUILTIN_CTYPE[dtype]
    raise ValueError(f"{dtype} is not a built-in data type")


def _normalize_type(dtype: DTypeLike) -> numpy.dtype[Any]:
    """
    Numpy's dtype shortcuts (e.g. ``numpy.int32``) are ``type`` objects
    and have slightly different properties from actual ``numpy.dtype`` objects.
    This function converts the former to ``numpy.dtype`` and keeps the latter unchanged.
    """
    return numpy.dtype(dtype)


def is_complex(dtype: DTypeLike) -> bool:
    """Returns ``True`` if ``dtype`` is complex."""
    dtype = _normalize_type(dtype)
    return numpy.issubdtype(dtype, numpy.complexfloating)


def is_double(dtype: DTypeLike) -> bool:
    """Returns ``True`` if ``dtype`` is double precision floating point."""
    dtype = _normalize_type(dtype)
    return numpy.issubdtype(dtype, numpy.float64) or numpy.issubdtype(dtype, numpy.complex128)


def is_integer(dtype: DTypeLike) -> bool:
    """Returns ``True`` if ``dtype`` is an integer."""
    dtype = _normalize_type(dtype)
    return numpy.issubdtype(dtype, numpy.integer)


def is_real(dtype: DTypeLike) -> bool:
    """Returns ``True`` if ``dtype`` is a real number (but not complex)."""
    dtype = _normalize_type(dtype)
    return numpy.issubdtype(dtype, numpy.floating)


def _promote_type(dtype: numpy.dtype[Any]) -> numpy.dtype[Any]:
    # not all numpy datatypes are supported by GPU, so we may need to promote
    if issubclass(dtype.type, numpy.signedinteger) and dtype.itemsize < 4:  # noqa: PLR2004
        return numpy.dtype("int32")
    if issubclass(dtype.type, numpy.unsignedinteger) and dtype.itemsize < 4:  # noqa: PLR2004
        return numpy.dtype("uint32")
    if issubclass(dtype.type, numpy.floating) and dtype.itemsize < 4:  # noqa: PLR2004
        return numpy.dtype("float32")
    # Not checking complex dtypes since there is nothing smaller than complex32, which is supported
    return dtype


def result_type(*dtypes: DTypeLike) -> numpy.dtype[Any]:
    """
    Wrapper for :py:func:`numpy.result_type`
    which takes into account types supported by GPUs.
    """
    return _promote_type(numpy.result_type(*dtypes))


def min_scalar_type(val: complex | numpy.number[Any]) -> numpy.dtype[Any]:
    """
    Wrapper for :py:func:`numpy.min_scalar_type`
    which takes into account types supported by GPUs.
    """
    dtype = val.dtype if isinstance(val, numpy.number) else numpy.min_scalar_type(val)
    return _promote_type(dtype)


def complex_for(dtype: DTypeLike) -> numpy.dtype[Any]:
    """Returns complex dtype corresponding to given floating point ``dtype``."""
    dtype = _normalize_type(dtype)
    if dtype == numpy.float32:
        return numpy.dtype("complex64")
    if dtype == numpy.float64:
        return numpy.dtype("complex128")
    raise ValueError(f"{dtype} does not have a corresponding complex type")


def real_for(dtype: DTypeLike) -> numpy.dtype[Any]:
    """Returns floating point dtype corresponding to given complex ``dtype``."""
    dtype = _normalize_type(dtype)
    if dtype == numpy.complex64:
        return numpy.dtype("float32")
    if dtype == numpy.complex128:
        return numpy.dtype("float64")
    raise ValueError(f"{dtype} does not have a corresponding real type")


def complex_ctr(dtype: DTypeLike) -> str:
    """Returns name of the constructor for the given ``dtype``."""
    return "COMPLEX_CTR(" + _ctype_builtin(_normalize_type(dtype)) + ")"


def _c_constant_arr(val: Any, shape: Sequence[int]) -> str:
    if len(shape) == 0:
        return c_constant(val)
    return "{" + ", ".join(_c_constant_arr(val[i], shape[1:]) for i in range(shape[0])) + "}"


def c_constant(
    val: complex | numpy.generic | numpy.ndarray[Any, numpy.dtype[Any]],
    dtype: DTypeLike | None = None,
) -> str:
    """
    Returns a C-style numerical constant.
    If ``val`` has a struct dtype, the generated constant will have the form ``{ ... }``
    and can be used as an initializer for a variable.
    """
    if dtype is not None:
        dtype = _promote_type(_normalize_type(dtype))
    elif isinstance(val, int | float | complex):
        dtype = min_scalar_type(val)
    else:
        dtype = _promote_type(val.dtype)

    numpy_val: numpy.generic | numpy.ndarray[Any, numpy.dtype[Any]]
    if isinstance(val, numpy.ndarray):
        numpy_val = numpy.asarray(val, dtype)
    else:
        numpy_val = numpy.asarray(val, dtype).flat[0]

    if len(numpy_val.shape) > 0:
        return _c_constant_arr(numpy_val, numpy_val.shape)

    scalar_val: numpy.generic
    scalar_val = numpy_val.flat[0] if isinstance(numpy_val, numpy.ndarray) else numpy_val

    if isinstance(scalar_val, numpy.void) and scalar_val.dtype.names is not None:
        return (
            "{" + ", ".join([c_constant(scalar_val[name]) for name in scalar_val.dtype.names]) + "}"
        )

    if isinstance(scalar_val, numpy.complexfloating):
        return (
            f"COMPLEX_CTR({_ctype_builtin(dtype)})"
            f"({c_constant(scalar_val.real)}, {c_constant(scalar_val.imag)})"
        )

    if isinstance(scalar_val, numpy.integer):
        if dtype.itemsize > 4:  # noqa: PLR2004
            postfix = "L" if numpy.issubdtype(scalar_val.dtype, numpy.signedinteger) else "UL"
        else:
            postfix = ""
        return str(scalar_val) + postfix

    if isinstance(scalar_val, numpy.floating):
        return repr(float(scalar_val)) + ("f" if scalar_val.dtype.itemsize <= 4 else "")  # noqa: PLR2004

    raise TypeError(f"Cannot render a value of type {type(val)} as a C constant")


def _struct_alignment(alignments: Iterable[int]) -> int:
    """
    Returns the minimum alignment for a structure given alignments for its fields.
    According to the C standard, it the lowest common multiple of the alignments
    of all of the members of the struct rounded up to the nearest power of two.
    """
    return bounding_power_of_2(_lcm(*alignments))


def _find_minimum_alignment(offset: int, base_alignment: int, prev_end: int) -> int:
    """
    Returns the minimum alignment that must be set for a field with
    ``base_alignment`` (the one inherent to the type),
    so that the compiler positioned it at ``offset`` given that the previous field
    ends at the position ``prev_end``.
    """
    # Essentially, we need to find the minimum k such that:
    # 1) offset = m * base_alignment * 2**k, where m > 0 and k >= 0;
    #    (by definition of alignment)
    # 2) offset - prev_offset < base_alignment * 2**k
    #    (otherwise the compiler can just as well take m' = m - 1).
    if offset % base_alignment != 0:
        raise ValueError(
            f"Field offset ({offset}) must be a multiple of the base alignment ({base_alignment})."
        )

    alignment = base_alignment
    while offset % alignment == 0:
        if offset - prev_end < alignment:
            return alignment

        alignment *= 2

    raise ValueError(
        f"Could not find a suitable alignment for the field at offset {offset}; "
        "consider adding explicit padding."
    )


class _WrappedType(NamedTuple):
    """Contains an accompanying information for an aligned dtype."""

    # The original dtype (note that it can be an array, that is have a non-empty ``shape``)
    dtype: numpy.dtype[Any]

    # The calculated structure alignment.
    # Can be different from ``dtype.alignment`` if it's needed to ensure the chosen ``itemsize``.
    alignment: int

    # A dictionary of `_WrappedType` objects for this dtype's fields.
    wrapped_fields: dict[str, _WrappedType]

    # An in integer if the type's alignment requires
    # an explicit specification in the C definition,
    # None otherwise.
    explicit_alignment: int | None

    # A dictionary of alignments for this dtype's fields;
    # similarly to `explicit_alignment`, a value is an integer if the alignment
    # has to be set explicitly, None otherwise.
    explicit_field_alignments: dict[str, int | None]

    @classmethod
    def _wrap_single(
        cls, dtype: numpy.dtype[Any], wrapped_fields: dict[str, _WrappedType]
    ) -> _WrappedType:
        """
        Attempts to build a `_WrappedType` object with the alignment information
        for the given dtype, aiming to preserve the given offsets and itemsizes.
        Raises a ``ValueError`` if it is not possible.
        """
        base_dtype = dtype.base

        # At this point dtype is guaranteed to be a struct type
        dtype_fields = cast(Mapping[str, tuple[numpy.dtype[Any], int]], base_dtype.fields)
        names = list(dtype_fields)

        # Find out what alignment has to be set for the field in order for the compiler
        # to place it at the offset specified in the description of `dtype`.
        explicit_field_alignments: dict[str, int | None] = {names[0]: None}
        field_alignments = [wrapped_fields[names[0]].alignment]
        for i in range(1, len(names)):
            name = names[i]
            prev_field_dtype, prev_offset = dtype_fields[names[i - 1]]
            _, offset = dtype_fields[name]
            prev_end = prev_offset + prev_field_dtype.itemsize
            inherent_alignment = wrapped_fields[name].alignment

            try:
                field_alignment = _find_minimum_alignment(offset, inherent_alignment, prev_end)
            except ValueError as exc:
                raise ValueError(f"Failed to find alignment for field `{name}`: {exc}") from exc

            explicit_field_alignments[name] = (
                field_alignment if field_alignment != inherent_alignment else None
            )
            field_alignments.append(field_alignment)

        # Same principle as above, but for the whole struct:
        # find out what alignment has to be set in order for the compiler
        # to place the next field at some dtype where this struct is a field type
        # at the offset corresponding to this struct's itemsize.

        last_dtype, last_offset = dtype_fields[names[-1]]
        struct_end = last_offset + last_dtype.itemsize

        # Find the total itemsize.
        # According to the standard, it must be a multiple of the struct alignment.
        base_struct_alignment = _struct_alignment(field_alignments)
        min_itemsize = min_blocks(struct_end, base_struct_alignment) * base_struct_alignment

        # A sanity check; `numpy` won't allow a dtype to have an itemsize smaller than the minimum
        assert base_dtype.itemsize >= min_itemsize, "dtype's itemsize is too small"  # noqa: S101

        if base_dtype.itemsize == min_itemsize:
            explicit_alignment = None
            alignment = base_struct_alignment
        elif base_dtype.itemsize != 2 ** log2(base_dtype.itemsize):
            raise ValueError("An itemsize that requires an explicit alignment must be a power of 2")
        else:
            explicit_alignment = base_dtype.itemsize
            alignment = explicit_alignment

        return cls(
            dtype=dtype,
            alignment=alignment,
            wrapped_fields=wrapped_fields,
            explicit_alignment=explicit_alignment,
            explicit_field_alignments=explicit_field_alignments,
        )

    @classmethod
    def _align_single(
        cls, dtype: numpy.dtype[Any], aligned_fields: dict[str, _WrappedType]
    ) -> _WrappedType:
        """
        Builds a `_WrappedType` object with the alignment information for a dtype,
        aligning it if it is not aligned, and checking the consistency of metadata if it is.
        """
        names = list(aligned_fields)

        # Build offsets for the structure using a procedure
        # similar to the one a compiler would use
        offsets = [0]
        for i in range(1, len(names)):
            prev_end = offsets[-1] + aligned_fields[names[i - 1]].dtype.itemsize
            alignment = aligned_fields[names[i]].alignment
            offsets.append(min_blocks(prev_end, alignment) * alignment)

        # Same principle as above, but for the whole struct:
        # find out the offset of the next struct of the same type,
        # which will be the struct's itemsize.
        # According to the standard, the itemsize must be a multiple of the struct alignment.
        field_alignments = [aligned_fields[name].alignment for name in names]
        struct_end = offsets[-1] + aligned_fields[names[-1]].dtype.itemsize
        base_struct_alignment = _struct_alignment(field_alignments)
        itemsize = min_blocks(struct_end, base_struct_alignment) * base_struct_alignment

        base_dtype = numpy.dtype(
            dict(
                names=list(aligned_fields),
                formats=[aligned.dtype for aligned in aligned_fields.values()],
                offsets=offsets,
                itemsize=itemsize,
                aligned=True,
            )
        )
        aligned_dtype = numpy.dtype((base_dtype, dtype.shape))

        # Now let the other method find out if any explicit alignments need to be set at this level.
        return cls._wrap_single(aligned_dtype, aligned_fields)

    @classmethod
    def wrap(cls, dtype: numpy.dtype[Any]) -> _WrappedType:
        if dtype.base.names is None:
            return cls(
                dtype=dtype,
                alignment=dtype.alignment,
                wrapped_fields={},
                explicit_alignment=None,
                explicit_field_alignments={},
            )

        # Since `.names` is not `None` at this point, we can restrict the type to help the inference
        dtype_fields = cast(Mapping[str, tuple[numpy.dtype[Any], int]], dtype.base.fields)

        wrapped_fields = {
            name: cls.wrap(field_dtype) for name, (field_dtype, _offset) in dtype_fields.items()
        }

        return cls._wrap_single(dtype, wrapped_fields)

    @classmethod
    def align(cls, dtype: numpy.dtype[Any]) -> _WrappedType:
        if dtype.base.names is None:
            return _WrappedType.wrap(dtype)

        # Since `.names` is not `None` at this point, we can restrict the type to help the inference
        dtype_fields = cast(Mapping[str, tuple[numpy.dtype[Any], int]], dtype.base.fields)

        aligned_fields = {
            name: cls.align(field_dtype) for name, (field_dtype, _offset) in dtype_fields.items()
        }

        return cls._align_single(dtype, aligned_fields)


def align(dtype: DTypeLike) -> numpy.dtype[Any]:
    """
    Returns a new struct dtype with the field offsets changed to the ones a compiler would use
    (without being given any explicit alignment qualifiers).
    Ignores all existing explicit itemsizes and offsets.
    """
    dtype = _normalize_type(dtype)
    wrapped_dtype = _WrappedType.align(dtype)
    return wrapped_dtype.dtype


def _lcm(*nums: int) -> int:
    """Returns the least common multiple of ``nums``."""
    if len(nums) == 1:
        return nums[0]
    if len(nums) == 2:  # noqa: PLR2004
        return nums[0] * nums[1] // gcd(nums[0], nums[1])
    return _lcm(nums[0], _lcm(*nums[1:]))


def _alignment_str(alignment: int | None) -> str:
    if alignment is not None:
        return "ALIGN(" + str(alignment) + ")"
    return ""


def _get_struct_module(wrapped_type: _WrappedType) -> Module:
    """
    Builds and returns a module with the C type definition for a given ``dtype``,
    possibly using modules for nested structures.
    """
    # `dtype.names` is not `None` at this point, restricting types
    dtype_fields = cast(Mapping[str, tuple[numpy.dtype[Any], int]], wrapped_type.dtype.base.fields)

    struct_alignment = wrapped_type.explicit_alignment
    field_alignments = wrapped_type.explicit_field_alignments

    # The tag (${prefix}_) is not necessary, but it helps to avoid
    # CUDA bug #1409907 (nested struct initialization like
    # "mystruct x = {0, {0, 0}, 0};" fails to compile)
    lines = ["typedef struct ${prefix}_ {"]
    kwds: dict[str, str | Module] = {}
    for name, (elem_dtype, _offset) in dtype_fields.items():
        base_elem_dtype = elem_dtype.base
        elem_dtype_shape = elem_dtype.shape

        array_suffix = "".join(f"[{d}]" for d in elem_dtype_shape)

        typename_var = "typename_" + name
        field_alignment = field_alignments[name]
        lines.append(
            f"    ${{{typename_var}}} {_alignment_str(field_alignment)} {name}{array_suffix};"
        )

        if base_elem_dtype.names is None:
            kwds[typename_var] = _ctype_builtin(base_elem_dtype)
        else:
            kwds[typename_var] = _get_struct_module(wrapped_type.wrapped_fields[name])

    lines.append("} " + _alignment_str(struct_alignment) + " ${prefix};")
    return Module.from_string("\n".join(lines), render_globals=kwds)


# TODO: this has a possibility of a memory leak, albeit it's very unlikely -
# the number of different types will most surely be limited.
# A way to avoid that is to introduce some custom class that encapsulates the dtype
# and the corresponding module (generated once on the first request).
# This way we can even have "newtypes" of sorts for dtypes with the same structure.
@functools.cache
def ctype(dtype: DTypeLike) -> str | Module:
    """
    Returns an object that can be passed as a global to :py:meth:`~grunnur.Program`
    and used to render a C equivalent of the given ``numpy`` dtype.
    If there is a built-in C equivalent, the object is just a string with the type name;
    otherwise it is a :py:class:`~grunnur.Module` object containing
    the corresponding ``struct`` declaration.

    The given ``dtype`` cannot be an array type (that is, its shape must be ``()``).

    Modules are cached, and the function returns a single module instance for equal ``dtype``'s.
    Therefore inside a kernel it will be rendered with the same prefix everywhere it is used.
    This results in a behavior characteristic for a structural type system,
    same as for the basic dtype-ctype conversion.

    .. note::

        If ``dtype`` is a struct type, it needs to be aligned
        (manually or with :py:func:`align`) in order for the field offsets
        on the host and the devices coincide.

    .. warning::

        ``numpy``'s ``isalignedstruct`` attribute, or ``aligned`` parameter
        is neither necessary nor sufficient for the memory layout
        on the host and the device to coincide.
        Use :py:func:`align` if you are not sure how to align your struct type.
    """
    # It doesn't seem like `numpy` is going to fix the alignment behavior.
    # For now the `isalignedstruct` attribute is more of an internal marker,
    # not useful for external users. So we will have to rely on `align()`.
    # See https://github.com/numpy/numpy/issues/4084 for the discussion.
    #
    # Would be nice to have our own marker, but it doesn't seem like `dtype`
    # can be safely inherited from.

    dtype = _normalize_type(dtype)

    if len(dtype.shape) > 0:
        raise ValueError("The data type cannot be an array")

    if dtype.names is None:
        return _ctype_builtin(dtype)

    wrapped_type = _WrappedType.wrap(dtype)
    return _get_struct_module(wrapped_type)


def _flatten_dtype(dtype: numpy.dtype[Any], offset: int, path: list[str | int]) -> list[FieldInfo]:
    if dtype.names is None:
        return [FieldInfo(path=path, dtype=dtype, offset=offset)]

    # `dtype.names` is not `None` at this point, restricting types
    dtype_fields = cast(Mapping[str, tuple[numpy.dtype[Any], int]], dtype.fields)

    result: list[FieldInfo] = []
    for name in dtype.names:
        elem_dtype, elem_offset = dtype_fields[name]

        elem_dtype_shape: tuple[int, ...]
        if len(elem_dtype.shape) == 0:
            base_elem_dtype = elem_dtype
            elem_dtype_shape = tuple()
        else:
            base_elem_dtype = elem_dtype.base
            elem_dtype_shape = elem_dtype.shape

        if len(elem_dtype_shape) == 0:
            result += _flatten_dtype(
                dtype=base_elem_dtype, offset=offset + elem_offset, path=[*path, name]
            )
        else:
            strides = list(
                reversed(numpy.multiply.accumulate([1, *reversed(elem_dtype_shape[1:])]))
            )
            for idxs in itertools.product(*[range(dim) for dim in elem_dtype_shape]):
                flat_idx = sum(idx * stride for idx, stride in zip(idxs, strides, strict=True))
                result += _flatten_dtype(
                    dtype=base_elem_dtype,
                    offset=offset + elem_offset + base_elem_dtype.itemsize * flat_idx,
                    path=[*path, name, *idxs],
                )
    return result


class FieldInfo(NamedTuple):
    """Metadata of a structure field."""

    path: list[str | int]
    """The full path to the field."""

    dtype: numpy.dtype[Any]
    """The field's datatype."""

    offset: int
    """The field's offset from the beginning of the struct."""

    @property
    def c_path(self) -> str:
        """Returns a string corresponding to the ``path`` to a struct element in C."""
        return "".join(
            (("." + elem) if isinstance(elem, str) else ("[" + str(elem) + "]"))
            for elem in self.path
        )


def flatten_dtype(dtype: DTypeLike) -> list[FieldInfo]:
    """
    Returns a list of tuples ``(path, dtype)`` for each of the basic dtypes in
    a (possibly nested) ``dtype``.
    ``path`` is a list of field names/array indices leading to the corresponding element.
    """
    dtype = _normalize_type(dtype)
    return _flatten_dtype(dtype, 0, [])


def _extract_field(
    arr: numpy.ndarray[Any, numpy.dtype[Any]], path: list[str | int], array_idxs: list[int]
) -> numpy.generic | numpy.ndarray[Any, numpy.dtype[Any]]:
    """
    A helper function for ``extract_field``.
    Need to collect array indices for dtype sub-array fields since they are attached to the end
    of the full array index.
    """
    if len(path) == 0:
        if len(array_idxs) == 0:
            return arr
        numpy_array_indices: list[slice | int] = [slice(None, None, None)] * (
            len(arr.shape) - len(array_idxs)
        )
        struct_array_indices: list[slice | int] = list(array_idxs)
        slices = tuple(numpy_array_indices + struct_array_indices)
        return arr[slices]

    if isinstance(path[0], str):
        return _extract_field(arr[path[0]], path[1:], array_idxs)

    return _extract_field(arr, path[1:], [*array_idxs, path[0]])


def extract_field(
    arr: numpy.ndarray[Any, numpy.dtype[Any]], path: list[str | int]
) -> numpy.generic | numpy.ndarray[Any, numpy.dtype[Any]]:
    """
    Extracts an element from an array of struct dtype.
    The ``path`` is the sequence of field names/array indices returned from
    :py:func:`~grunnur.dtypes.flatten_dtype`.
    """
    return _extract_field(arr, path, [])



from __future__ import annotations

from tempfile import mkdtemp
from typing import TYPE_CHECKING, Any

import numpy

# Skipping coverage count - can't test properly
try:  # pragma: no cover
    import pycuda.compiler as pycuda_compiler
    import pycuda.driver as pycuda_driver
    import pycuda.tools as pycuda_tools
except ImportError:  # pragma: no cover
    # these variables are used for a PyCUDA mock during tests
    pycuda_driver = None  # type: ignore[assignment]
    pycuda_compiler = None  # type: ignore[assignment]
    pycuda_tools = None  # type: ignore[assignment]

from . import dtypes
from .adapter_base import (
    APIID,
    AdapterCompilationError,
    APIAdapter,
    APIAdapterFactory,
    BufferAdapter,
    ContextAdapter,
    DeviceAdapter,
    DeviceParameters,
    DeviceType,
    KernelAdapter,
    PlatformAdapter,
    PreparedKernelAdapter,
    ProgramAdapter,
    QueueAdapter,
)
from .array_metadata import ArrayMetadataLike
from .template import Template
from .utils import get_launch_size, normalize_object_sequence, prod

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Mapping, Sequence


# Another way would be to place it in the try block and only set `_avaialable`
# if CUDA was initialized correctly, but it is better to distinguish import errors
# (PyCUDA not installed, which is quite common) and initialization errors
# (some problem with the package).
#
# Skipping coverage count - can't test properly
if pycuda_driver is not None:  # pragma: no cover
    pycuda_driver.init()


_API_ID = APIID("cuda")
_TEMPLATE = Template.from_associated_file(__file__)
_PRELUDE = _TEMPLATE.get_def("prelude")
_CONSTANT_ARRAYS_DEF = _TEMPLATE.get_def("constant_arrays_def")


class CuAPIAdapterFactory(APIAdapterFactory):
    @property
    def api_id(self) -> APIID:
        return _API_ID

    @property
    def available(self) -> bool:
        return pycuda_driver is not None

    def make_api_adapter(self) -> CuAPIAdapter:
        if not self.available:
            raise ImportError(
                "CUDA API is not operational. Check if PyCUDA is installed correctly."
            )

        return CuAPIAdapter()


class CuAPIAdapter(APIAdapter):
    @property
    def id(self) -> APIID:
        return _API_ID

    @property
    def platform_count(self) -> int:
        return 1

    def get_platform_adapters(self) -> tuple[CuPlatformAdapter, ...]:
        return (CuPlatformAdapter(self),)

    def isa_backend_device(self, obj: Any) -> bool:
        return isinstance(obj, pycuda_driver.Device)

    def isa_backend_platform(self, _obj: Any) -> bool:
        return False  # CUDA backend doesn't have platforms

    def isa_backend_context(self, obj: Any) -> bool:
        return isinstance(obj, pycuda_driver.Context)

    def make_device_adapter(self, pycuda_device: pycuda_driver.Device) -> CuDeviceAdapter:
        return CuDeviceAdapter.from_pycuda_device(pycuda_device)

    def make_platform_adapter(self, _pycuda_platform: Any) -> CuPlatformAdapter:  # pragma: no cover
        # Not going to be called since `isa_backend_platform()` always returns `False`
        raise RuntimeError("CUDA does not have the concept of a platform")

    def make_context_adapter_from_backend_contexts(
        self, backend_contexts: Sequence[Any], *, take_ownership: bool
    ) -> CuContextAdapter:
        return CuContextAdapter.from_pycuda_contexts(
            backend_contexts, take_ownership=take_ownership
        )

    def make_context_adapter_from_device_adapters(
        self, device_adapters: Sequence[DeviceAdapter]
    ) -> CuContextAdapter:
        cu_device_adapters = normalize_object_sequence(device_adapters, CuDeviceAdapter)
        return CuContextAdapter.from_device_adapters(cu_device_adapters)


class CuPlatformAdapter(PlatformAdapter):
    def __init__(self, api_adapter: CuAPIAdapter):
        self._api_adapter = api_adapter

    def __eq__(self, other: object) -> bool:
        return isinstance(other, CuPlatformAdapter) and self._api_adapter == other._api_adapter

    def __hash__(self) -> int:
        return hash((type(self), self._api_adapter))

    @property
    def api_adapter(self) -> CuAPIAdapter:
        return self._api_adapter

    @property
    def platform_idx(self) -> int:
        return 0

    @property
    def name(self) -> str:
        return "nVidia CUDA"

    @property
    def vendor(self) -> str:
        return "nVidia"

    @property
    def version(self) -> str:
        return "CUDA " + ".".join(str(x) for x in pycuda_driver.get_version())

    @property
    def device_count(self) -> int:
        return pycuda_driver.Device.count()

    def get_device_adapters(self) -> tuple[CuDeviceAdapter, ...]:
        return tuple(
            CuDeviceAdapter(self, pycuda_driver.Device(device_idx), device_idx)
            for device_idx in range(self.device_count)
        )


class CuDeviceAdapter(DeviceAdapter):
    @classmethod
    def from_pycuda_device(cls, pycuda_device: pycuda_driver.Device) -> CuDeviceAdapter:
        """Creates this object from a PyCuda ``Device`` object."""
        platform_adapter = CuPlatformAdapter(CuAPIAdapter())
        for device_idx, device_adapter in enumerate(platform_adapter.get_device_adapters()):
            if pycuda_device == device_adapter.pycuda_device:
                return cls(platform_adapter, pycuda_device, device_idx)

        # Sanity check, should not be reachable as long as `pycuda` is consistent.
        raise RuntimeError(f"{pycuda_device} was not found among CUDA devices")  # pragma: no cover

    def __init__(
        self,
        platform_adapter: CuPlatformAdapter,
        pycuda_device: pycuda_driver.Device,
        device_idx: int,
    ):
        self._platform_adapter = platform_adapter
        self._device_idx = device_idx
        self.pycuda_device = pycuda_device
        self._params: CuDeviceParameters | None = None

    def __eq__(self, other: object) -> bool:
        return isinstance(other, CuDeviceAdapter) and self.pycuda_device == other.pycuda_device

    def __hash__(self) -> int:
        return hash((type(self), self.pycuda_device))

    @property
    def platform_adapter(self) -> CuPlatformAdapter:
        return self._platform_adapter

    @property
    def device_idx(self) -> int:
        return self._device_idx

    @property
    def name(self) -> str:
        return self.pycuda_device.name()

    @property
    def params(self) -> CuDeviceParameters:
        if self._params is None:
            self._params = CuDeviceParameters(self.pycuda_device)
        return self._params


class CuDeviceParameters(DeviceParameters):
    def __init__(self, pycuda_device: pycuda_driver.Device):
        self._type = DeviceType.GPU

        self._max_total_local_size = pycuda_device.max_threads_per_block
        self._max_local_sizes = (
            pycuda_device.max_block_dim_x,
            pycuda_device.max_block_dim_y,
            pycuda_device.max_block_dim_z,
        )

        self._max_num_groups = (
            pycuda_device.max_grid_dim_x,
            pycuda_device.max_grid_dim_y,
            pycuda_device.max_grid_dim_z,
        )

        # there is no corresponding constant in the API at the moment
        self._local_mem_banks = 16 if pycuda_device.compute_capability()[0] < 2 else 32  # noqa: PLR2004

        self._warp_size = pycuda_device.warp_size
        self._local_mem_size = pycuda_device.max_shared_memory_per_block
        self._compute_units = pycuda_device.multiprocessor_count
        self._devdata = pycuda_tools.DeviceData(pycuda_device)

    @property
    def max_total_local_size(self) -> int:
        return self._max_total_local_size

    @property
    def max_local_sizes(self) -> tuple[int, ...]:
        return self._max_local_sizes

    @property
    def warp_size(self) -> int:
        return self._warp_size

    @property
    def max_num_groups(self) -> tuple[int, ...]:
        return self._max_num_groups

    @property
    def type(self) -> DeviceType:
        return self._type

    @property
    def local_mem_size(self) -> int:
        return self._local_mem_size

    @property
    def local_mem_banks(self) -> int:
        return self._local_mem_banks

    @property
    def compute_units(self) -> int:
        return self._compute_units

    def align_words(self, word_size: int) -> int:
        return self._devdata.align_words(word_size)


class _ContextStack:
    """A helper class that keeps the CUDA context stack state."""

    def __init__(self, pycuda_contexts: Sequence[pycuda_driver.Context], *, take_ownership: bool):
        self._active_context: int | None = None
        self._owns_contexts = take_ownership
        self._pycuda_contexts: dict[int, pycuda_driver.Context] = {}

        self.device_order: list[int] = []
        self.device_adapters: dict[int, CuDeviceAdapter] = {}

        # A shortcut for the case when we're given a single external context
        # and told not to take ownership.
        dont_push = len(pycuda_contexts) == 1 and not take_ownership

        for context in pycuda_contexts:
            if not dont_push:
                context.push()
            device = context.get_device()
            if not dont_push:
                context.pop()

            device_adapter = CuDeviceAdapter.from_pycuda_device(device)
            self.device_adapters[device_adapter.device_idx] = device_adapter
            self._pycuda_contexts[device_adapter.device_idx] = context
            self.device_order.append(device_adapter.device_idx)

    def deactivate(self) -> None:
        if self._active_context is not None and self._owns_contexts:
            self._pycuda_contexts[self._active_context].pop()
            self._active_context = None

    def activate(self, device_idx: int) -> None:
        if self._active_context != device_idx and self._owns_contexts:
            self.deactivate()
            self._pycuda_contexts[device_idx].push()
            self._active_context = device_idx

    def __del__(self) -> None:
        self.deactivate()


class CuContextAdapter(ContextAdapter):
    """Wraps CUDA contexts for several devices."""

    @classmethod
    def from_pycuda_devices(
        cls, pycuda_devices: Sequence[pycuda_driver.Device]
    ) -> CuContextAdapter:
        """Creates a context based on one or several (distinct) PyCuda ``Device`` objects."""
        pycuda_devices = normalize_object_sequence(pycuda_devices, pycuda_driver.Device)
        contexts = []
        for device in pycuda_devices:
            context = device.make_context()
            pycuda_driver.Context.pop()
            contexts.append(context)
        return cls(contexts, take_ownership=True)

    @classmethod
    def from_pycuda_contexts(
        cls, pycuda_contexts: Sequence[pycuda_driver.Context], *, take_ownership: bool
    ) -> CuContextAdapter:
        """
        Creates a context based on one or several (distinct) PyCuda ``Context`` objects.
        If ``take_ownership`` is ``False``, and there is only one context given,
        it should be pushed to the top of the context stack.
        Otherwise, none of the PyCuda contexts should be pushed to the context stack.
        """
        if len(pycuda_contexts) > 1 and not take_ownership:
            raise ValueError(
                "When dealing with multiple CUDA contexts, Grunnur must be the one managing them"
            )
        pycuda_contexts = normalize_object_sequence(pycuda_contexts, pycuda_driver.Context)
        return cls(pycuda_contexts, take_ownership=take_ownership)

    @classmethod
    def from_device_adapters(cls, device_adapters: Sequence[CuDeviceAdapter]) -> CuContextAdapter:
        """
        Creates a context based on one or several (distinct)
        :py:class:`CuDeviceAdapter` objects.
        """
        return cls.from_pycuda_devices(
            [device_adapter.pycuda_device for device_adapter in device_adapters]
        )

    def __init__(self, pycuda_contexts: Sequence[pycuda_driver.Context], *, take_ownership: bool):
        self._context_stack = _ContextStack(pycuda_contexts, take_ownership=take_ownership)
        self._device_adapters = self._context_stack.device_adapters
        self._device_order = self._context_stack.device_order

        # Eager activation to help some tests.
        self._context_stack.activate(next(iter(self._device_order)))

    @property
    def device_adapters(self) -> Mapping[int, CuDeviceAdapter]:
        return self._device_adapters

    @property
    def device_order(self) -> list[int]:
        return self._device_order

    def activate_device(self, device_idx: int) -> None:
        """
        Activates the device ``device_idx``.
        Pops a previous context from the stack, if there was one pushed before,
        and pushes the corresponding context to the stack.
        """
        self._context_stack.activate(device_idx)

    def deactivate(self) -> None:
        self._context_stack.deactivate()

    @staticmethod
    def render_prelude(*, fast_math: bool = False) -> str:
        return _PRELUDE.render(fast_math=fast_math)

    def compile_single_device(
        self,
        device_adapter: DeviceAdapter,
        prelude: str,
        src: str,
        *,
        keep: bool = False,
        fast_math: bool = False,
        compiler_options: Sequence[str] | None = None,
        constant_arrays: Mapping[str, ArrayMetadataLike] | None = None,
    ) -> CuProgramAdapter:
        # Will be checked in the higher levels
        assert isinstance(device_adapter, CuDeviceAdapter)  # noqa: S101

        normalized_constant_arrays = normalize_constant_arrays(constant_arrays)
        constant_arrays_src = _CONSTANT_ARRAYS_DEF.render(
            dtypes=dtypes, constant_arrays=normalized_constant_arrays
        )

        if compiler_options is None:
            compiler_options = []

        options = list(compiler_options) + (["-use_fast_math"] if fast_math else [])
        full_src = prelude + constant_arrays_src + src
        self.activate_device(device_adapter.device_idx)

        # For some reason, `keep=True` does not work without an explicit `cache_dir`.
        # A new temporary dir is still created, and everything is placed there,
        # and `cache_dir` receives a copy of `kernel.cubin`.
        cache_dir = mkdtemp() if keep else None

        try:
            module = pycuda_compiler.SourceModule(
                full_src, no_extern_c=True, options=options, keep=keep, cache_dir=cache_dir
            )
        except pycuda_driver.CompileError as exc:
            raise AdapterCompilationError(exc, full_src) from exc

        return CuProgramAdapter(self, device_adapter.device_idx, module, full_src)

    def allocate(self, device_adapter: DeviceAdapter, size: int) -> CuBufferAdapter:
        # Will be checked in the higher levels
        assert isinstance(device_adapter, CuDeviceAdapter)  # noqa: S101
        return CuBufferAdapter(self, device_adapter.device_idx, size)

    def make_queue_adapter(self, device_adapter: DeviceAdapter) -> CuQueueAdapter:
        # Will be checked in the higher levels
        assert isinstance(device_adapter, CuDeviceAdapter)  # noqa: S101
        self.activate_device(device_adapter.device_idx)
        stream = pycuda_driver.Stream()
        return CuQueueAdapter(self, device_adapter.device_idx, stream)


class CuBufferAdapter(BufferAdapter):
    def __init__(
        self,
        context_adapter: CuContextAdapter,
        device_idx: int,
        size: int,
        *,
        offset: int = 0,
        managed: bool = False,  # noqa: ARG002
        base_buffer: CuBufferAdapter | None = None,
    ):
        if base_buffer is None:
            context_adapter.activate_device(device_idx)
            base_allocation = pycuda_driver.mem_alloc(size)
        else:
            base_allocation = base_buffer._base_allocation  # noqa: SLF001

        self._size = size
        self._offset = offset
        self._context_adapter = context_adapter
        self._device_idx = device_idx

        self._base_buffer = base_buffer
        self._base_allocation: pycuda_driver.DeviceAllocation = base_allocation
        self._ptr = numpy.uintp(self._base_allocation) + numpy.uintp(offset)

    @property
    def kernel_arg(self) -> numpy.uintp:
        return self._ptr

    @property
    def size(self) -> int:
        return self._size

    @property
    def offset(self) -> int:
        return self._offset

    def get_sub_region(self, origin: int, size: int) -> CuBufferAdapter:
        # Sanity check, should be already checked in `Buffer`
        assert origin + size <= self._size  # noqa: S101
        base_buffer = self if self._base_buffer is None else self._base_buffer
        return CuBufferAdapter(
            self._context_adapter,
            self._device_idx,
            size,
            offset=self._offset + origin,
            base_buffer=base_buffer,
        )

    def set(
        self,
        queue_adapter: QueueAdapter,
        source: numpy.ndarray[Any, numpy.dtype[Any]] | BufferAdapter,
        *,
        no_async: bool = False,
    ) -> None:
        # Will be checked in the upper levels.
        assert isinstance(queue_adapter, CuQueueAdapter)  # noqa: S101
        device_idx = queue_adapter._device_idx  # noqa: SLF001
        self._context_adapter.activate_device(device_idx)

        # PyCUDA needs pointers to be passed as `numpy.number` to kernels,
        # but `memcpy` functions require Python `int`s.
        ptr = int(self._ptr) if isinstance(self._ptr, numpy.number) else self._ptr

        if isinstance(source, numpy.ndarray):
            if no_async:
                pycuda_driver.memcpy_htod(ptr, source)
            else:
                pycuda_driver.memcpy_htod_async(ptr, source, stream=queue_adapter._pycuda_stream)  # noqa: SLF001
        else:
            # Will be checked in the upper levels.
            assert isinstance(source, CuBufferAdapter)  # noqa: S101
            buf_ptr = int(source._ptr) if isinstance(source._ptr, numpy.number) else source._ptr  # noqa: SLF001
            if no_async:
                pycuda_driver.memcpy_dtod(ptr, buf_ptr, source.size)
            else:
                pycuda_driver.memcpy_dtod_async(
                    ptr,
                    buf_ptr,
                    source.size,
                    stream=queue_adapter._pycuda_stream,  # noqa: SLF001
                )

    def get(
        self,
        queue_adapter: QueueAdapter,
        host_array: numpy.ndarray[Any, numpy.dtype[Any]],
        *,
        async_: bool = False,
    ) -> None:
        # Will be checked in the upper levels.
        assert isinstance(queue_adapter, CuQueueAdapter)  # noqa: S101
        device_idx = queue_adapter._device_idx  # noqa: SLF001
        self._context_adapter.activate_device(device_idx)

        # PyCUDA needs pointers to be passed as `numpy.number` to kernels,
        # but `memcpy` functions require Python `int`s.
        ptr = int(self._ptr) if isinstance(self._ptr, numpy.number) else self._ptr

        if async_:
            pycuda_driver.memcpy_dtoh_async(host_array, ptr, stream=queue_adapter._pycuda_stream)  # noqa: SLF001
        else:
            pycuda_driver.memcpy_dtoh(host_array, ptr)


def normalize_constant_arrays(
    constant_arrays: Mapping[str, ArrayMetadataLike] | None,
) -> dict[str, tuple[int, numpy.dtype[Any]]]:
    if constant_arrays is None:
        constant_arrays = {}

    normalized = {}
    for name, metadata in constant_arrays.items():
        if not isinstance(metadata, ArrayMetadataLike):
            raise TypeError("Unknown constant array metadata type")
        dtype = metadata.dtype
        length = prod(metadata.shape)
        normalized[name] = (length, dtype)

    return normalized


class CuQueueAdapter(QueueAdapter):
    def __init__(
        self,
        context_adapter: CuContextAdapter,
        device_idx: int,
        pycuda_stream: pycuda_driver.Stream,
    ):
        self._context_adapter = context_adapter
        self._device_idx = device_idx
        self._pycuda_stream = pycuda_stream

    def synchronize(self) -> None:
        self._context_adapter.activate_device(self._device_idx)
        self._pycuda_stream.synchronize()


class CuProgramAdapter(ProgramAdapter):
    def __init__(
        self,
        context_adapter: CuContextAdapter,
        device_idx: int,
        pycuda_program: pycuda_compiler.SourceModule,
        source: str,
    ):
        self._context_adapter = context_adapter
        self._device_idx = device_idx
        self._pycuda_program = pycuda_program
        self._source = source

    @property
    def source(self) -> str:
        return self._source

    def __getattr__(self, kernel_name: str) -> CuKernelAdapter:
        pycuda_kernel = self._pycuda_program.get_function(kernel_name)
        return CuKernelAdapter(self, self._device_idx, pycuda_kernel)

    def set_constant_buffer(
        self,
        queue_adapter: QueueAdapter,
        name: str,
        arr: BufferAdapter | numpy.ndarray[Any, numpy.dtype[Any]],
    ) -> None:
        """Uploads a constant array ``arr`` corresponding to the symbol ``name`` to the context."""
        # Will be checked in the upper levels
        assert isinstance(queue_adapter, CuQueueAdapter)  # noqa: S101

        self._context_adapter.activate_device(self._device_idx)
        symbol, size = self._pycuda_program.get_global(name)

        pycuda_stream = queue_adapter._pycuda_stream  # noqa: SLF001

        if isinstance(arr, BufferAdapter):
            # Will be checked in the upper levels
            assert isinstance(arr, CuBufferAdapter)  # noqa: S101
            transfer_size = arr.size
        elif isinstance(arr, numpy.ndarray):
            transfer_size = prod(arr.shape) * arr.dtype.itemsize
        else:  # pragma: no cover
            # Shouldn't reach this path because the type is already checked by the caller.
            # Nevertheless leaving it here as a sanity check.
            raise TypeError(f"Unsupported array type: {type(arr)}")

        if transfer_size != size:
            raise ValueError(
                f"Incorrect size of the constant buffer; "
                f"expected {size} bytes, got {transfer_size}"
            )

        if isinstance(arr, CuBufferAdapter):
            pycuda_driver.memcpy_dtod_async(symbol, int(arr._ptr), arr.size, stream=pycuda_stream)  # noqa: SLF001
        else:
            # This serves two purposes:
            # 1. Gives us a pagelocked array, as PyCUDA requires
            # 2. Makes the array contiguous
            # Constant array are usually quite small, so it won't affect the performance.
            buf = pycuda_driver.pagelocked_empty(arr.shape, arr.dtype)
            numpy.copyto(buf, arr)
            pycuda_driver.memcpy_htod_async(symbol, buf, stream=pycuda_stream)


class CuKernelAdapter(KernelAdapter):
    def __init__(
        self,
        program_adapter: CuProgramAdapter,
        device_idx: int,
        pycuda_function: pycuda_compiler.Function,
    ):
        self._context_adapter = program_adapter._context_adapter  # noqa: SLF001
        self._program_adapter = program_adapter
        self._device_idx = device_idx
        self._pycuda_function = pycuda_function

    @property
    def program_adapter(self) -> CuProgramAdapter:
        return self._program_adapter

    @property
    def max_total_local_size(self) -> int:
        return self._pycuda_function.get_attribute(
            pycuda_driver.function_attribute.MAX_THREADS_PER_BLOCK
        )

    def prepare(
        self, global_size: Sequence[int], local_size: Sequence[int] | None = None
    ) -> CuPreparedKernelAdapter:
        return CuPreparedKernelAdapter(self, global_size, local_size)


class CuPreparedKernelAdapter(PreparedKernelAdapter):
    def __init__(
        self,
        kernel_adapter: CuKernelAdapter,
        global_size: Sequence[int],
        local_size: Sequence[int] | None = None,
    ):
        context_adapter = kernel_adapter._context_adapter  # noqa: SLF001
        device_idx = kernel_adapter._device_idx  # noqa: SLF001

        device_adapter = context_adapter._device_adapters[device_idx]  # noqa: SLF001

        grid, block = get_launch_size(
            device_adapter.params.max_local_sizes,
            device_adapter.params.max_total_local_size,
            global_size,
            local_size,
        )

        # append missing dimensions, otherwise PyCUDA will complain
        max_dims = len(device_adapter.params.max_local_sizes)
        self._block = block + (1,) * (max_dims - len(block))
        self._grid = grid + (1,) * (max_dims - len(grid))

        self._device_idx = device_idx
        self._context_adapter = context_adapter
        self._pycuda_function = kernel_adapter._pycuda_function  # noqa: SLF001

    def __call__(
        self,
        queue_adapter: QueueAdapter,
        *args: BufferAdapter | numpy.generic,
        local_mem: int = 0,
    ) -> None:
        # Will be checked in the upper levels
        assert isinstance(queue_adapter, CuQueueAdapter)  # noqa: S101

        # Sanity check. If it happened, there's something wrong in the abstraction layer logic
        # (Program and PreparedKernel).
        assert self._device_idx == queue_adapter._device_idx  # noqa: SLF001, S101

        backend_args = []
        for arg in args:
            if isinstance(arg, BufferAdapter):
                karg = arg.kernel_arg
                backend_args.append(karg)
            else:
                backend_args.append(arg)

        self._context_adapter.activate_device(self._device_idx)
        return self._pycuda_function(
            *backend_args,
            grid=self._grid,
            block=self._block,
            stream=queue_adapter._pycuda_stream,  # noqa: SLF001
            shared=local_mem,
        )



import numpy

from grunnur import API, dtypes
from grunnur.device_discovery import select_devices


def get_test_array(shape, dtype, *, strides=None, offset=0, no_zeros=False, high=None):
    dtype = numpy.dtype(dtype)  # normalize to access the fields

    rng = numpy.random.default_rng()

    if offset != 0:
        raise NotImplementedError

    if dtype.names is not None:
        result = numpy.empty(shape, dtype)
        for name in dtype.names:
            result[name] = get_test_array(shape, dtype[name], no_zeros=no_zeros, high=high)
    else:
        if dtypes.is_integer(dtype):
            low = 1 if no_zeros else 0
            if high is None:
                high = 100  # will work even with signed chars
            get_arr = lambda: rng.integers(low, high, shape, dtype)
        else:
            low = 0.01 if no_zeros else 0
            if high is None:
                high = 1.0
            get_arr = lambda: rng.uniform(low, high, shape).astype(dtype)

        result = (get_arr() + 1j * get_arr()) if dtypes.is_complex(dtype) else get_arr()

    if strides is not None:
        result = numpy.lib.stride_tricks.as_strided(result, result.shape, strides)

    return result


def check_select_devices(
    mock_stdin,
    mock_backend_factory,
    capsys,
    platforms_devices,
    *,
    inputs=None,
    interactive=False,
    quantity=1,
    device_filter=None,
    platform_filter=None,
):
    # CUDA API has a single fixed platform, so using the OpenCL one
    backend = mock_backend_factory.mock_pyopencl()

    for platform_name, device_infos in platforms_devices:
        backend.add_platform_with_devices(platform_name, device_infos)

    if inputs is not None:
        for line in inputs:
            mock_stdin.line(line)

    api = API.from_api_id(backend.api_id)

    try:
        devices = select_devices(
            api,
            interactive=interactive,
            quantity=quantity,
            device_filter=device_filter,
            platform_filter=platform_filter,
        )
        assert mock_stdin.empty()
    finally:
        # Otherwise the output will be shown in the console
        capsys.readouterr()

    return devices



from __future__ import annotations

import weakref
from abc import ABC, abstractmethod
from collections import Counter
from collections.abc import Iterable, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    NamedTuple,
    Protocol,
    cast,
    runtime_checkable,
)
from weakref import ReferenceType

from .adapter_base import BufferAdapter, QueueAdapter
from .array import Array
from .buffer import Buffer
from .sorted_list import SortedList

if TYPE_CHECKING:  # pragma: no cover
    import numpy

    from .context import BoundDevice
    from .queue import Queue


@runtime_checkable
class VirtualAllocations(Protocol):
    def __virtual_allocations__(self) -> Sequence[Array | Buffer]:  # pragma: no cover
        ...


def extract_dependencies(dependencies: Any) -> set[int]:
    """Recursively extracts allocation identifiers from an iterable or an ``Array`` object."""
    if isinstance(dependencies, VirtualBufferAdapter):
        return {dependencies._id}  # noqa: SLF001
    if isinstance(dependencies, Buffer):
        return extract_dependencies(dependencies._buffer_adapter)  # noqa: SLF001
    if isinstance(dependencies, Array):
        return extract_dependencies(dependencies.data)
    if isinstance(dependencies, Iterable):
        results = set()
        for dep in dependencies:
            results.update(extract_dependencies(dep))
        return results
    if isinstance(dependencies, VirtualAllocations):
        # a hook for exposing nested virtual allocations in arbitrary classes
        return extract_dependencies(dependencies.__virtual_allocations__)

    return set()


class VirtualAllocator:
    """
    A helper callable object to use as an allocator
    for :py:class:`~grunnur.Array` creation.
    Encapsulates the dependencies (as identifiers, doesn't hold references for actual objects).
    """

    def __init__(self, manager: VirtualManager, dependencies: set[int]):
        self.manager = manager
        self.dependencies = dependencies

    def __call__(self, device: BoundDevice, size: int) -> Buffer:
        # TODO: seems redundant to pass a device here,
        # but it must mimic the API of ``Buffer.allocate()``.
        if device != self.manager.device:
            raise ValueError(
                f"This allocator is attached to device {self.manager.device}, "
                f"but was asked to allocate on {device}"
            )
        return self.manager._allocate_virtual(size, self.dependencies)  # noqa: SLF001


class VirtualBufferAdapter(BufferAdapter):
    """A virtual buffer object."""

    def __init__(self, manager: VirtualManager, size: int, id_: int, buffer_adapter: BufferAdapter):
        self._manager = manager
        self._id = id_
        self._size = size
        self._real_buffer_adapter = buffer_adapter

    @property
    def kernel_arg(self) -> Any:
        return self._real_buffer_adapter.kernel_arg

    def get_sub_region(self, origin: int, size: int) -> BufferAdapter:
        # TODO: how to handle it?
        raise NotImplementedError("Virtual buffers do not support subregions")

    def get(
        self,
        queue_adapter: QueueAdapter,
        host_array: numpy.ndarray[Any, numpy.dtype[Any]],
        *,
        async_: bool = False,
    ) -> None:
        return self._real_buffer_adapter.get(queue_adapter, host_array, async_=async_)

    def set(
        self,
        queue_adapter: QueueAdapter,
        source: numpy.ndarray[Any, numpy.dtype[Any]] | BufferAdapter,
        *,
        no_async: bool = False,
    ) -> None:
        return self._real_buffer_adapter.set(queue_adapter, source, no_async=no_async)

    @property
    def offset(self) -> int:
        return 0

    @property
    def size(self) -> int:
        return self._size

    def _set_real_buffer_adapter(self, buf: Buffer) -> None:
        self._real_buffer_adapter = buf._buffer_adapter  # noqa: SLF001


class VirtualManager(ABC):
    """
    Base class for a manager of virtual allocations.

    :param context: an instance of :py:class:`~grunnur.Context`.
    """

    def __init__(self, device: BoundDevice):
        self.device = device
        self._id_counter = 0
        self._virtual_buffers: dict[int, ReferenceType[VirtualBufferAdapter]] = {}

    def allocator(self, dependencies: Any | None = None) -> VirtualAllocator:
        """
        Create a callable to use for :py:class:`~grunnur.Array` creation.

        :param dependencies: can be a :py:class:`~grunnur.Array` instance
            (the ones containing persistent allocations will be ignored),
            an iterable with valid values,
            or an object with the attribute ``__virtual_allocations__`` which is a valid value
            (the last two will be processed recursively).
        """
        dependencies = extract_dependencies(dependencies)
        return VirtualAllocator(self, dependencies)

    def _allocate_virtual(self, size: int, dependencies: set[int]) -> Buffer:
        new_id = self._id_counter
        self._id_counter += 1

        if not dependencies.issubset(self._virtual_buffers):
            missing_deps = dependencies.difference(self._virtual_buffers)
            missing_deps_str = ", ".join(str(dep) for dep in missing_deps)
            raise ValueError(
                f"Some of the declared dependencies (with IDs {missing_deps_str}) do not exist"
            )

        rbuf = self._allocate_specific(new_id, size, dependencies)
        vbuf = VirtualBufferAdapter(self, size, new_id, rbuf._buffer_adapter)  # noqa: SLF001
        self._virtual_buffers[new_id] = weakref.ref(vbuf, lambda _: self._free(new_id))

        return Buffer(self.device, vbuf)

    def _update_buffer(self, id_: int) -> None:
        vbuf = self._virtual_buffers[id_]()
        # Sanity check - this shouldn't happen by construction
        assert vbuf is not None  # noqa: S101
        buf = self._get_real_buffer(id_)
        vbuf._set_real_buffer_adapter(buf)  # noqa: SLF001

    def _update_all(self) -> None:
        for id_ in self._virtual_buffers:
            self._update_buffer(id_)

    def _free(self, id_: int) -> None:
        del self._virtual_buffers[id_]
        self._free_specific(id_)

    def pack(self, queue: Queue) -> None:
        """
        Packs the real allocations possibly reducing total memory usage.
        This process can be slow and may synchronize the base queue.
        """
        self._pack_specific(queue)
        self._update_all()

    def statistics(self) -> VirtualAllocationStatistics:
        """Returns allocation statistics."""
        return VirtualAllocationStatistics(
            self._real_buffers(),
            # cast() to override inference here - we know that vb() will not return `None`
            [cast(VirtualBufferAdapter, vb()) for vb in self._virtual_buffers.values()],
        )

    @abstractmethod
    def _allocate_specific(self, new_id: int, size: int, dependencies: set[int]) -> Buffer:
        pass

    @abstractmethod
    def _get_real_buffer(self, id_: int) -> Buffer:
        pass

    @abstractmethod
    def _real_buffers(self) -> list[Buffer]:
        pass

    @abstractmethod
    def _free_specific(self, id_: int) -> None:
        pass

    @abstractmethod
    def _pack_specific(self, queue: Queue) -> None:
        pass


class VirtualAllocationStatistics:
    """Virtual allocation details."""

    real_size_total: int
    """The total size of physical allocations (in bytes)."""

    real_num: int
    """The number of physical allocations."""

    real_sizes: dict[int, int]
    """A dictionary ``size: count`` with the counts for physical allocations of each size."""

    virtual_size_total: int
    """The total size of virtual allocations (in bytes)."""

    virtual_num: int
    """The number of virtual allocations."""

    virtual_sizes: dict[int, int]
    """A dictionary ``size: count`` with the counts for virtual allocations of each size."""

    def __init__(
        self, real_buffers: Iterable[Buffer], virtual_buffers: Iterable[VirtualBufferAdapter]
    ):
        real_sizes = [rb.size for rb in real_buffers]
        virtual_sizes = [vb.size for vb in virtual_buffers]

        self.real_size_total: int = sum(real_sizes)
        self.real_num: int = len(real_sizes)
        self.real_sizes: dict[int, int] = dict(Counter(real_sizes))

        self.virtual_size_total: int = sum(virtual_sizes)
        self.virtual_num: int = len(virtual_sizes)
        self.virtual_sizes: dict[int, int] = dict(Counter(virtual_sizes))

    def __str__(self) -> str:
        real_buffers = ", ".join(f"{num}x{size}b" for size, num in sorted(self.real_sizes.items()))
        virtual_buffers = ", ".join(
            f"{num}x{size}b" for size, num in sorted(self.virtual_sizes.items())
        )
        return (
            f"VirtualAllocationStatistics("
            f"real: {self.real_num} allocs, "
            f"total size {self.real_size_total}b, "
            f"buffers: {real_buffers}; "
            f"virtual: {self.virtual_num} allocs, "
            f"total size {self.virtual_size_total}b, "
            f"buffers: {virtual_buffers})"
        )


class TrivialManager(VirtualManager):
    """Trivial manager --- allocates a separate buffer for each allocation request."""

    def __init__(self, device: BoundDevice):
        VirtualManager.__init__(self, device)
        self._rbuffers: dict[int, Buffer] = {}

    def _allocate_specific(self, new_id: int, size: int, _dependencies: set[int]) -> Buffer:
        buf = Buffer.allocate(self.device, size)
        self._rbuffers[new_id] = buf
        return buf

    def _get_real_buffer(self, id_: int) -> Buffer:
        return self._rbuffers[id_]

    def _free_specific(self, id_: int) -> None:
        del self._rbuffers[id_]

    def _pack_specific(self, queue: Queue) -> None:
        pass

    def _real_buffers(self) -> list[Buffer]:
        return list(self._rbuffers.values())


class ZeroOffsetManager(VirtualManager):
    """
    Tries to assign several allocation requests to a single real allocation,
    if dependencies allow that.
    All virtual allocations start from the beginning of real allocations.
    """

    class VirtualAllocation(NamedTuple):
        size: int
        dependencies: set[int]

    class RealAllocation(NamedTuple):
        buffer: Buffer
        virtual_ids: set[int]

    class RealSize(NamedTuple):
        size: int
        real_id: int

    class VirtualMapping(NamedTuple):
        real_id: int
        sub_region: Buffer

    def __init__(self, device: BoundDevice):
        VirtualManager.__init__(self, device)

        self._virtual_allocations: dict[int, ZeroOffsetManager.VirtualAllocation] = {}
        self._real_sizes = SortedList[ZeroOffsetManager.RealSize]((), key=lambda x: x.size)
        self._virtual_to_real: dict[int, ZeroOffsetManager.VirtualMapping] = {}
        self._real_allocations: dict[int, ZeroOffsetManager.RealAllocation] = {}
        self._real_id_counter = 0

    def _allocate_specific(self, new_id: int, size: int, dependencies: set[int]) -> Buffer:
        # Dependencies should be bidirectional.
        # So if some new allocation says it depends on earlier ones,
        # we need to update their dependency lists.
        for dep in dependencies:
            self._virtual_allocations[dep].dependencies.add(new_id)

        # Save virtual allocation parameters
        self._virtual_allocations[new_id] = self.VirtualAllocation(size, dependencies)

        # Find a real allocation using the greedy algorithm.
        return self._fast_add(new_id, size, dependencies)

    def _fast_add(self, new_id: int, size: int, dependencies: set[int]) -> Buffer:
        """Greedy algorithm to find a real allocation for a given virtual allocation."""
        # Find the smallest real allocation which can hold the requested virtual allocation.
        try:
            idx_start = self._real_sizes.argfind_ge(size)
        except ValueError:
            idx_start = len(self._real_sizes)

        # Check all real allocations with suitable sizes, starting from the smallest one.
        # Use the first real allocation which does not contain ``new_id``'s dependencies.
        for idx in range(idx_start, len(self._real_sizes)):
            real_id = self._real_sizes[idx].real_id
            buf = self._real_allocations[real_id].buffer
            virtual_ids = self._real_allocations[real_id].virtual_ids
            if virtual_ids.isdisjoint(dependencies):
                virtual_ids.add(new_id)
                break
        else:
            # If no suitable real allocation is found, create a new one.
            buf = Buffer.allocate(self.device, size)
            real_id = self._real_id_counter
            self._real_id_counter += 1

            self._real_allocations[real_id] = self.RealAllocation(buf, {new_id})
            self._real_sizes.insert(self.RealSize(size, real_id))

        # TODO: Here it would be more appropriate to use buffer.get_sub_region(0, size),
        # but OpenCL does not allow several overlapping subregions to be used in a single kernel
        # for both read and write, which ruins the whole idea.
        # So we are passing full buffers and hope that the Array class takes care of sizes.
        self._virtual_to_real[new_id] = self.VirtualMapping(
            real_id, self._real_allocations[real_id].buffer
        )

        return buf

    def _get_real_buffer(self, id_: int) -> Buffer:
        return self._virtual_to_real[id_].sub_region

    def _free_specific(self, id_: int) -> None:
        # Remove the allocation from the dependency lists of its dependencies
        dependencies = self._virtual_allocations[id_].dependencies
        for dep in dependencies:
            self._virtual_allocations[dep].dependencies.remove(id_)

        vtr = self._virtual_to_real[id_]

        # Clear virtual allocation data
        del self._virtual_allocations[id_]
        del self._virtual_to_real[id_]

        # Fast and non-optimal free.
        # Remove the virtual allocation from the real allocation,
        # and delete the real allocation if its no longer used by other virtual allocations.
        ra = self._real_allocations[vtr.real_id]
        ra.virtual_ids.remove(id_)
        if len(ra.virtual_ids) == 0:
            del self._real_allocations[vtr.real_id]
            self._real_sizes.remove(self.RealSize(ra.buffer.size, vtr.real_id))

    def _pack_specific(self, queue: Queue) -> None:
        """
        Full memory re-pack.
        In theory, should find the optimal (with the minimal real allocation size) distribution
        of virtual allocations.
        """
        # Need to synchronize, because we are going to change allocation addresses,
        # and we do not want to free the memory some kernel is reading from.
        queue.synchronize()

        # Clear all real allocation data.
        self._real_sizes = SortedList((), key=lambda x: x.size)
        self._real_allocations = {}
        self._real_id_counter = 0

        va = self._virtual_allocations

        # Sort all virtual allocations by size
        virtual_sizes = sorted([(va[id_].size, id_) for id_ in va], key=lambda x: x[0])

        # Application of greedy algorithm for virtual allocations starting from the largest one
        # should give the optimal distribution.
        for size, id_ in reversed(virtual_sizes):
            self._fast_add(id_, size, self._virtual_allocations[id_].dependencies)

    def _real_buffers(self) -> list[Buffer]:
        return [ra.buffer for ra in self._real_allocations.values()]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import importlib.metadata
import os

import setuptools_scm

# -- Project information -----------------------------------------------------

project = "grunnur"
copyright = "2020–now, Bogdan Opanchuk"
author = "Bogdan Opanchuk"

# The full version, including alpha/beta/rc tags
try:
    release = importlib.metadata.version(project)
except importlib.metadata.PackageNotFoundError:
    release = setuptools_scm.get_version(relative_to=os.path.abspath("../pyproject.toml"))


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
]

autoclass_content = "both"
autodoc_member_order = "groupwise"
autodoc_type_aliases = dict(DTypeLike="DTypeLike")

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Note: set to the lower bound of `numpy` version in the dependencies;
# must be kept synchronized.
intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/1.22", None),
    "python": ("https://docs.python.org/3", None),
    "mako": ("https://docs.makotemplates.org/en/latest/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []



from __future__ import annotations

import os
from pathlib import Path
from tempfile import mkdtemp
from typing import TYPE_CHECKING, Any, cast

# Skipping coverage count - can't test properly
try:  # pragma: no cover
    import pyopencl
except ImportError:  # pragma: no cover
    pyopencl = None  # type: ignore[assignment]

from . import dtypes
from .adapter_base import (
    APIID,
    AdapterCompilationError,
    APIAdapter,
    APIAdapterFactory,
    BufferAdapter,
    ContextAdapter,
    DeviceAdapter,
    DeviceParameters,
    DeviceType,
    KernelAdapter,
    PlatformAdapter,
    PreparedKernelAdapter,
    ProgramAdapter,
    QueueAdapter,
)
from .template import Template
from .utils import normalize_object_sequence

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Mapping, Sequence

    import numpy

    from .array_metadata import ArrayMetadataLike


_API_ID = APIID("opencl")
_TEMPLATE = Template.from_associated_file(__file__)
_PRELUDE = _TEMPLATE.get_def("prelude")


class OclAPIAdapterFactory(APIAdapterFactory):
    @property
    def api_id(self) -> APIID:
        return _API_ID

    @property
    def available(self) -> bool:
        return pyopencl is not None

    def make_api_adapter(self) -> OclAPIAdapter:
        if not self.available:
            raise ImportError(
                "OpenCL API is not operational. Check if PyOpenCL is installed correctly."
            )

        return OclAPIAdapter()


class OclAPIAdapter(APIAdapter):
    @property
    def id(self) -> APIID:
        return _API_ID

    @property
    def platform_count(self) -> int:
        return len(pyopencl.get_platforms())

    def get_platform_adapters(self) -> tuple[OclPlatformAdapter, ...]:
        return tuple(
            OclPlatformAdapter(self, platform, platform_idx)
            for platform_idx, platform in enumerate(pyopencl.get_platforms())
        )

    def isa_backend_device(self, obj: Any) -> bool:
        return isinstance(obj, pyopencl.Device)

    def isa_backend_platform(self, obj: Any) -> bool:
        return isinstance(obj, pyopencl.Platform)

    def isa_backend_context(self, obj: Any) -> bool:
        return isinstance(obj, pyopencl.Context)

    def make_device_adapter(self, pyopencl_device: pyopencl.Device) -> OclDeviceAdapter:
        return OclDeviceAdapter.from_pyopencl_device(pyopencl_device)

    def make_platform_adapter(self, pyopencl_platform: pyopencl.Platform) -> OclPlatformAdapter:
        return OclPlatformAdapter.from_pyopencl_platform(pyopencl_platform)

    def make_context_adapter_from_device_adapters(
        self, device_adapters: Sequence[DeviceAdapter]
    ) -> OclContextAdapter:
        ocl_device_adapters = normalize_object_sequence(device_adapters, OclDeviceAdapter)
        return OclContextAdapter.from_device_adapters(ocl_device_adapters)

    def make_context_adapter_from_backend_contexts(
        self,
        pyopencl_contexts: Sequence[Any],
        *,
        take_ownership: bool,  # noqa: ARG002
    ) -> OclContextAdapter:
        if len(pyopencl_contexts) > 1:
            raise ValueError("Cannot make one OpenCL context out of several contexts")
        return OclContextAdapter.from_pyopencl_context(pyopencl_contexts[0])


class OclPlatformAdapter(PlatformAdapter):
    @classmethod
    def from_pyopencl_platform(cls, pyopencl_platform: pyopencl.Platform) -> OclPlatformAdapter:
        """Creates this object from a PyOpenCL ``Platform`` object."""
        api_adapter = OclAPIAdapter()
        platform_adapters = api_adapter.get_platform_adapters()
        for platform_idx, platform_adapter in enumerate(platform_adapters):
            if pyopencl_platform == platform_adapter._pyopencl_platform:  # noqa: SLF001
                return cls(api_adapter, pyopencl_platform, platform_idx)

        # Sanity check, should not be reachable as long as `pyopencl` is consistent.
        raise RuntimeError(
            f"{pyopencl_platform} was not found among OpenCL platforms"
        )  # pragma: no cover

    def __init__(
        self, api_adapter: APIAdapter, pyopencl_platform: pyopencl.Platform, platform_idx: int
    ):
        self._api_adapter = api_adapter
        self._pyopencl_platform = pyopencl_platform
        self._platform_idx = platform_idx

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, OclPlatformAdapter)
            and self._pyopencl_platform == other._pyopencl_platform
        )

    def __hash__(self) -> int:
        return hash((type(self), self._pyopencl_platform))

    @property
    def api_adapter(self) -> APIAdapter:
        return self._api_adapter

    @property
    def platform_idx(self) -> int:
        return self._platform_idx

    @property
    def name(self) -> str:
        return self._pyopencl_platform.name

    @property
    def vendor(self) -> str:
        return self._pyopencl_platform.vendor

    @property
    def version(self) -> str:
        return self._pyopencl_platform.version

    @property
    def device_count(self) -> int:
        return len(self._pyopencl_platform.get_devices())

    def get_device_adapters(self) -> tuple[OclDeviceAdapter, ...]:
        return tuple(
            OclDeviceAdapter(self, device, device_idx)
            for device_idx, device in enumerate(self._pyopencl_platform.get_devices())
        )


class OclDeviceAdapter(DeviceAdapter):
    @classmethod
    def from_pyopencl_device(cls, pyopencl_device: pyopencl.Device) -> OclDeviceAdapter:
        """Creates this object from a PyOpenCL ``Device`` object."""
        platform_adapter = OclPlatformAdapter.from_pyopencl_platform(pyopencl_device.platform)
        for device_idx, device_adapter in enumerate(platform_adapter.get_device_adapters()):
            if pyopencl_device == device_adapter._pyopencl_device:  # noqa: SLF001
                return cls(platform_adapter, pyopencl_device, device_idx)

        # Sanity check, should not be reachable as long as `pyopencl` is consistent.
        raise RuntimeError(
            f"{pyopencl_device} was not found among OpenCL devices"
        )  # pragma: no cover

    def __init__(
        self, platform_adapter: PlatformAdapter, pyopencl_device: pyopencl.Device, device_idx: int
    ):
        self._platform_adapter = platform_adapter
        self._device_idx = device_idx
        self._pyopencl_device = pyopencl_device
        self._params: OclDeviceParameters | None = None

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, OclDeviceAdapter) and self._pyopencl_device == other._pyopencl_device
        )

    def __hash__(self) -> int:
        return hash((type(self), self._pyopencl_device))

    @property
    def platform_adapter(self) -> PlatformAdapter:
        return self._platform_adapter

    @property
    def device_idx(self) -> int:
        return self._device_idx

    @property
    def name(self) -> str:
        return self._pyopencl_device.name

    @property
    def params(self) -> DeviceParameters:
        if self._params is None:
            self._params = OclDeviceParameters(self._pyopencl_device)
        return self._params


class OclDeviceParameters(DeviceParameters):
    def __init__(self, pyopencl_device: pyopencl.Device):
        self._type = (
            DeviceType.CPU if pyopencl_device.type == pyopencl.device_type.CPU else DeviceType.GPU
        )

        if (
            pyopencl_device.platform.name == "Apple"
            and pyopencl_device.type == pyopencl.device_type.CPU
        ):
            self._max_local_sizes: tuple[int, ...]

            # Apple is being funny again.
            # As of MacOS 10.15.0, it reports the maximum local size as 1024 (for the device),
            # when even for the simplest kernel it turns out to be 128.
            # Moreover, if local_barrier() is used in the kernel, it becomes 1.
            self._max_total_local_size = 1
            self._max_local_sizes = (1, 1, 1)
        else:
            self._max_total_local_size = pyopencl_device.max_work_group_size
            self._max_local_sizes = tuple(pyopencl_device.max_work_item_sizes)

        # The limit of `2**address_bits` is for global size
        max_size = 2**pyopencl_device.address_bits // self._max_total_local_size
        self._max_num_groups = (max_size, max_size, max_size)

        if pyopencl_device.type == pyopencl.device_type.CPU:
            # For CPU both values do not make much sense
            self._local_mem_banks = 1
            self._warp_size = 1
        elif "cl_nv_device_attribute_query" in pyopencl_device.extensions:
            # If NV extensions are available, use them to query info
            cc = getattr(pyopencl_device, "compute_capability_major_nv", 1)
            self._local_mem_banks = 16 if cc < 2 else 32  # noqa: PLR2004
            self._warp_size = pyopencl_device.warp_size_nv
        elif pyopencl_device.vendor == "NVIDIA":
            # nVidia device, but no extensions.
            # Must be APPLE OpenCL implementation.
            # Assuming CC>=3
            self._local_mem_banks = 32
            self._warp_size = 32
        else:
            # AMD card.
            # Do not know how to query this info, so settle for most probable values.

            self._local_mem_banks = 32

            # An alternative is to query CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE
            # for some arbitrary kernel.
            self._warp_size = 32

        self._local_mem_size = pyopencl_device.local_mem_size
        self._compute_units = pyopencl_device.max_compute_units

    @property
    def max_total_local_size(self) -> int:
        return self._max_total_local_size

    @property
    def max_local_sizes(self) -> tuple[int, ...]:
        return self._max_local_sizes

    @property
    def warp_size(self) -> int:
        return self._warp_size

    @property
    def max_num_groups(self) -> tuple[int, ...]:
        return self._max_num_groups

    @property
    def type(self) -> DeviceType:
        return self._type

    @property
    def local_mem_size(self) -> int:
        return self._local_mem_size

    @property
    def local_mem_banks(self) -> int:
        return self._local_mem_banks

    @property
    def compute_units(self) -> int:
        return self._compute_units

    def align_words(self, word_size: int) -> int:
        # TODO: is it possible to query it from the API?
        return {4: 16, 8: 16, 16: 8}[word_size]


class OclContextAdapter(ContextAdapter):
    @classmethod
    def from_pyopencl_devices(
        cls, pyopencl_devices: Sequence[pyopencl.Device]
    ) -> OclContextAdapter:
        """Creates a context based on one or several (distinct) PyOpenCL ``Device`` objects."""
        pyopencl_devices = normalize_object_sequence(pyopencl_devices, pyopencl.Device)
        return cls(pyopencl.Context(pyopencl_devices), pyopencl_devices=pyopencl_devices)

    @classmethod
    def from_pyopencl_context(cls, pyopencl_context: pyopencl.Context) -> OclContextAdapter:
        """Creates a context based on a (possibly multi-device) PyOpenCL ``Context`` object."""
        return cls(pyopencl_context)

    @classmethod
    def from_device_adapters(cls, device_adapters: Sequence[OclDeviceAdapter]) -> OclContextAdapter:
        """
        Creates a context based on one or several (distinct)
        :py:class:`OclDeviceAdapter` objects.
        """
        return cls.from_pyopencl_devices(
            [device_adapter._pyopencl_device for device_adapter in device_adapters]  # noqa: SLF001
        )

    def __init__(
        self,
        pyopencl_context: pyopencl.Context,
        pyopencl_devices: Sequence[pyopencl.Device] | None = None,
    ):
        if pyopencl_devices is None:
            pyopencl_devices = pyopencl_context.devices

        # Not checking here that all the `pyopencl_devices` are actually present
        # in `pyopencl_context`, since the constructor should only be called
        # from the trusted classmethods, which already ensure that.

        self._pyopencl_context = pyopencl_context

        device_adapters = [
            OclDeviceAdapter.from_pyopencl_device(device) for device in pyopencl_devices
        ]

        self._device_order = [device_adapter.device_idx for device_adapter in device_adapters]

        self._pyopencl_devices = dict(zip(self._device_order, pyopencl_devices, strict=True))

        self._device_adapters = {
            device_adapter.device_idx: device_adapter for device_adapter in device_adapters
        }

        self.platform_adapter = device_adapters[0].platform_adapter

        # On an Apple platform, in a multi-device context with nVidia cards it is necessary to have
        # any created OpenCL buffers allocated on the host (with ALLOC_HOST_PTR flag).
        # If it is not the case, using a subregion of such a buffer leads to a crash.
        is_multi_device = len(pyopencl_devices) > 1
        is_apple_platform = pyopencl_devices[0].platform.name == "Apple"
        has_geforce_device = any("GeForce" in device.name for device in pyopencl_devices)
        self._buffers_host_allocation = is_multi_device and is_apple_platform and has_geforce_device

    @property
    def device_adapters(self) -> Mapping[int, DeviceAdapter]:
        return self._device_adapters

    @property
    def device_order(self) -> list[int]:
        return self._device_order

    @staticmethod
    def render_prelude(*, fast_math: bool = False) -> str:
        return _PRELUDE.render(fast_math=fast_math, dtypes=dtypes)

    def compile_single_device(
        self,
        device_adapter: DeviceAdapter,
        prelude: str,
        src: str,
        *,
        keep: bool = False,
        fast_math: bool = False,
        compiler_options: Sequence[str] | None = None,
        constant_arrays: Mapping[str, ArrayMetadataLike] | None = None,
    ) -> OclProgramAdapter:
        # Will be checked in the upper levels.
        assert isinstance(device_adapter, OclDeviceAdapter)  # noqa: S101

        # Sanity check: should have been caught in compile()
        assert not constant_arrays  # noqa: S101

        src = prelude + src

        if keep:
            temp_dir = Path(mkdtemp())
            temp_file_path = temp_dir / "kernel.cl"

            with temp_file_path.open("w") as f:
                # `str()` is for convenience of mocking;
                # during a normal operation `src` is already a string
                f.write(str(src))

            print("*** compiler output in", temp_dir)  # noqa: T201

        else:
            temp_dir = None

        if compiler_options is None:
            compiler_options = []

        options = list(compiler_options) + (
            ["-cl-mad-enable", "-cl-fast-relaxed-math"] if fast_math else []
        )

        pyopencl_program = pyopencl.Program(self._pyopencl_context, src)

        try:
            pyopencl_program.build(
                devices=[self._pyopencl_devices[device_adapter.device_idx]],
                options=options,
                cache_dir=str(temp_dir) if temp_dir else None,
            )
        except pyopencl.RuntimeError as exc:
            raise AdapterCompilationError(exc, src) from exc

        return OclProgramAdapter(self, device_adapter, pyopencl_program, src)

    def allocate(self, device_adapter: DeviceAdapter, size: int) -> OclBufferAdapter:
        # Will be checked in the upper levels.
        assert isinstance(device_adapter, OclDeviceAdapter)  # noqa: S101

        flags = pyopencl.mem_flags.READ_WRITE
        if self._buffers_host_allocation:
            flags |= pyopencl.mem_flags.ALLOC_HOST_PTR

        pyopencl_buffer = pyopencl.Buffer(self._pyopencl_context, flags, size=size)

        return OclBufferAdapter(self, device_adapter, pyopencl_buffer)

    def make_queue_adapter(self, device_adapter: DeviceAdapter) -> OclQueueAdapter:
        # Will be checked in the upper levels.
        assert isinstance(device_adapter, OclDeviceAdapter)  # noqa: S101
        queue = pyopencl.CommandQueue(
            self._pyopencl_context,
            device=self._device_adapters[device_adapter.device_idx]._pyopencl_device,  # noqa: SLF001
        )

        return OclQueueAdapter(self, device_adapter, queue)

    def deactivate(self) -> None:
        # No need to do anything for OpenCL
        pass


class OclBufferAdapter(BufferAdapter):
    def __init__(
        self,
        context_adapter: OclContextAdapter,
        device_adapter: OclDeviceAdapter,
        pyopencl_buffer: pyopencl.Buffer,
    ):
        self._context_adapter = context_adapter
        self._pyopencl_buffer = pyopencl_buffer
        self._device_adapter = device_adapter

    @property
    def kernel_arg(self) -> pyopencl.Buffer:
        return self._pyopencl_buffer

    @property
    def size(self) -> int:
        return self._pyopencl_buffer.size

    @property
    def offset(self) -> int:
        return self._pyopencl_buffer.offset

    def set(
        self,
        queue_adapter: QueueAdapter,
        source: numpy.ndarray[Any, numpy.dtype[Any]] | BufferAdapter,
        *,
        no_async: bool = False,
    ) -> None:
        # Will be checked in the upper levels.
        assert isinstance(queue_adapter, OclQueueAdapter)  # noqa: S101
        buf_data: pyopencl.Buffer | numpy.ndarray[Any, numpy.dtype[Any]]
        if isinstance(source, BufferAdapter):
            # Will be checked in the upper levels.
            assert isinstance(source, OclBufferAdapter)  # noqa: S101
            buf_data = source._pyopencl_buffer  # noqa: SLF001
        else:
            buf_data = source

        # This keyword is only supported for transfers involving hosts in PyOpenCL
        kwds = {}
        if not isinstance(source, OclBufferAdapter):
            kwds["is_blocking"] = no_async

        pyopencl.enqueue_copy(
            queue_adapter._pyopencl_queue,  # noqa: SLF001
            self._pyopencl_buffer,
            buf_data,
            **kwds,
        )

    def get(
        self,
        queue_adapter: QueueAdapter,
        host_array: numpy.ndarray[Any, numpy.dtype[Any]],
        *,
        async_: bool = False,
    ) -> None:
        # Will be checked in the upper levels.
        assert isinstance(queue_adapter, OclQueueAdapter)  # noqa: S101
        pyopencl.enqueue_copy(
            queue_adapter._pyopencl_queue,  # noqa: SLF001
            host_array,
            self._pyopencl_buffer,
            is_blocking=not async_,
        )

    def get_sub_region(self, origin: int, size: int) -> OclBufferAdapter:
        return OclBufferAdapter(
            self._context_adapter,
            self._device_adapter,
            self._pyopencl_buffer.get_sub_region(origin, size),
        )


class OclQueueAdapter(QueueAdapter):
    def __init__(
        self,
        context_adapter: OclContextAdapter,
        device_adapter: OclDeviceAdapter,
        pyopencl_queue: pyopencl.CommandQueue,
    ):
        self._context_adapter = context_adapter
        self._device_adapter = device_adapter
        self._pyopencl_queue = pyopencl_queue

    def synchronize(self) -> None:
        self._pyopencl_queue.finish()


class OclProgramAdapter(ProgramAdapter):
    def __init__(
        self,
        context_adapter: OclContextAdapter,
        device_adapter: OclDeviceAdapter,
        pyopencl_program: pyopencl.Program,
        source: str,
    ):
        self._context_adapter = context_adapter
        self._device_adapter = device_adapter
        self._pyopencl_program = pyopencl_program
        self._source = source

    @property
    def source(self) -> str:
        return self._source

    def __getattr__(self, kernel_name: str) -> OclKernelAdapter:
        pyopencl_kernel = getattr(self._pyopencl_program, kernel_name)
        return OclKernelAdapter(self, self._device_adapter, pyopencl_kernel)

    def set_constant_buffer(
        self,
        _queue_adapter: QueueAdapter,
        _name: str,
        _arr: BufferAdapter | numpy.ndarray[Any, numpy.dtype[Any]],
    ) -> None:
        raise RuntimeError("OpenCL does not allow setting constant arrays externally")


class OclKernelAdapter(KernelAdapter):
    def __init__(
        self,
        program_adapter: OclProgramAdapter,
        device_adapter: OclDeviceAdapter,
        pyopencl_kernel: pyopencl.Kernel,
    ):
        self._program_adapter = program_adapter
        self._device_adapter = device_adapter
        self._pyopencl_kernel = pyopencl_kernel

    @property
    def program_adapter(self) -> OclProgramAdapter:
        return self._program_adapter

    def prepare(
        self, global_size: Sequence[int], local_size: Sequence[int] | None = None
    ) -> OclPreparedKernelAdapter:
        return OclPreparedKernelAdapter(self, global_size, local_size)

    @property
    def max_total_local_size(self) -> int:
        return cast(
            int,
            self._pyopencl_kernel.get_work_group_info(
                pyopencl.kernel_work_group_info.WORK_GROUP_SIZE,
                self._device_adapter._pyopencl_device,  # noqa: SLF001
            ),
        )


class OclPreparedKernelAdapter(PreparedKernelAdapter):
    def __init__(
        self,
        kernel_adapter: OclKernelAdapter,
        global_size: Sequence[int],
        local_size: Sequence[int] | None = None,
    ):
        self._kernel_adapter = kernel_adapter
        self._local_size = local_size
        self._global_size = global_size
        self._device_adapter = kernel_adapter._device_adapter  # noqa: SLF001
        self._pyopencl_kernel = kernel_adapter._pyopencl_kernel  # noqa: SLF001

    def __call__(
        self,
        queue_adapter: QueueAdapter,
        *args: BufferAdapter | numpy.generic,
        local_mem: int = 0,
    ) -> pyopencl.Event:
        # Local memory size is passed via regular kernel arguments in OpenCL.
        # Should be checked in `PreparedKernel`.
        assert local_mem == 0  # noqa: S101

        # We have to keep the signature more general because of the base class,
        # but the upper levels will ensure this is the case.
        assert isinstance(queue_adapter, OclQueueAdapter)  # noqa: S101

        # Sanity check. If it happened, there's something wrong in the abstraction layer logic
        # (Program and PreparedKernel).
        assert self._device_adapter == queue_adapter._device_adapter  # noqa: S101, SLF001

        backend_args: list[pyopencl.Buffer | numpy.generic] = []
        for arg in args:
            if isinstance(arg, BufferAdapter):
                karg = arg.kernel_arg
                # Need this to pass it to PyOpenCL. It may be an actual Buffer, or a virtual buffer
                assert isinstance(karg, pyopencl.Buffer)  # noqa: S101
                backend_args.append(karg)
            else:
                backend_args.append(arg)

        return self._pyopencl_kernel(
            queue_adapter._pyopencl_queue,  # noqa: SLF001
            self._global_size,
            self._local_size,
            *backend_args,
        )



from collections.abc import Sequence
from enum import Enum, Flag
from typing import Any

import numpy

class Platform:
    name: str
    vendor: str
    version: str

    def get_devices(self) -> list[Device]: ...

class device_type(Enum):  # noqa: N801
    CPU = ...
    GPU = ...

class Device:
    platform: Platform
    name: str
    type: device_type
    max_work_group_size: int
    max_work_item_sizes: list[int]
    address_bits: int
    extensions: str
    compute_capability_major_nv: int
    warp_size_nv: int
    vendor: str
    local_mem_size: int
    max_compute_units: int

class Context:
    devices: list[Device]

    def __init__(self, devices: Sequence[Device]): ...

class Program:
    def __init__(self, context: Context, src: str): ...
    def build(
        self,
        options: Sequence[str] = [],
        devices: Sequence[Device] | None = None,
        cache_dir: str | None = None,
    ) -> None: ...

class kernel_work_group_info:  # noqa: N801
    WORK_GROUP_SIZE: Any

class Kernel:
    def __call__(
        self,
        queue: CommandQueue,
        global_size: Sequence[int],
        local_size: Sequence[int] | None,
        *args: Buffer | numpy.generic,
    ) -> Event: ...
    def get_work_group_info(self, param: kernel_work_group_info, device: Device) -> Any: ...

class Event: ...

def get_platforms() -> list[Platform]: ...

# That's what it's called in PyOpenCL, can't help it.
class RuntimeError(Exception): ...  # noqa: A001

class mem_flags(Flag):  # noqa: N801
    _NONE = ...
    READ_WRITE = ...
    ALLOC_HOST_PTR = ...

class Buffer:
    size: int
    offset: int
    def __init__(self, context: Context, flags: mem_flags = mem_flags._NONE, size: int = 0): ...  # noqa: PYI011, SLF001
    def get_sub_region(
        self,
        origin: int,
        size: int,
        flags: mem_flags = mem_flags._NONE,  # noqa: PYI011, SLF001
    ) -> Buffer: ...

class CommandQueue:
    def __init__(self, context: Context, device: Device | None = None): ...
    def finish(self) -> None: ...

def enqueue_copy(
    queue: CommandQueue,
    dest: Buffer | numpy.ndarray[Any, numpy.dtype[Any]],
    src: Buffer | numpy.ndarray[Any, numpy.dtype[Any]],
    *,
    is_blocking: bool = True,
) -> None: ...
